# #############################################################################
# sv_deconvnet.py
# ======================
# Authors :
# Eric BEZZAM [ebezzam@gmail.com]
# #############################################################################

import numpy as np
import torch
from lensless.recon.trainable_recon import TrainableReconstructionAlgorithm


def compute_weight_matrices(psf, K):
    Nx, Ny = psf.shape[1:3]

    # get centers of KxK patches
    centers = []
    for i in range(K):
        for j in range(K):
            centers.append((int((i + 0.5) * Nx / K), int((j + 0.5) * Ny / K)))

    # compute weight matrices
    Y, X = np.meshgrid(np.arange(Ny), np.arange(Nx))
    eps = 1e-4
    weight_mat = []
    for center in centers:
        weight = ((X - center[0]) ** 2 + (Y - center[1]) ** 2 + eps) ** (-0.5)
        weight_mat.append(weight)

    # normalize weight matrices
    sum_weights = np.sum(weight_mat, axis=0)
    for i in range(K * K):
        weight_mat[i] /= sum_weights

    # check that sums to 1 at each pixel
    sum_mat = np.sum(weight_mat, axis=0)
    assert np.allclose(sum_mat, 1.0)

    return weight_mat


class SVDeconvNet(TrainableReconstructionAlgorithm):
    def __init__(self, psf, dtype=None, K=3, **kwargs):
        """
        Constructor for SVDeconvNet as proposed in PhoCoLens: https://phocolens.github.io/

        Parameters
        ----------
        psf : :py:class:`~torch.Tensor`
            Point spread function (PSF) that models forward propagation.
            Must be of shape (depth, height, width, channels) even if
            depth = 1 and channels = 1. You can use :py:func:`~lensless.io.load_psf`
            to load a PSF from a file such that it is in the correct format.
        dtype : float32 or float64
            Data type to use for optimization.
        K : int
            (K x K) kernels are learned for spatially-variant deconvolution.

        """
        multipsf = psf.repeat(K * K, 1, 1, 1)

        # compute weight matrices
        self.weight_mat = torch.tensor(compute_weight_matrices(psf, K)).to(
            dtype=psf.dtype, device=psf.device
        )
        self.weight_mat = self.weight_mat[None, :, :, :, None]  # (batch, K*K, Nx, Ny, channels)
        super(SVDeconvNet, self).__init__(multipsf, n_iter=1, dtype=dtype, reset=False, **kwargs)
        self.reset()

    def _form_image(self):
        self._image_est[self._image_est < 0] = 0
        return self._image_est

    def _set_psf(self, psf):
        super()._set_psf(psf)

    def reset(self, batch_size=1):
        # no state variables
        return

    def _update(self, iter):
        self._image_est = torch.sum(
            self.weight_mat * self._convolver.deconvolve(self._data), dim=1, keepdim=True
        )



# #############################################################################
# utils.py
# =================
# Authors :
# Yohann PERRON [yohann.perron@gmail.com]
# Eric BEZZAM [ebezzam@gmail.com]
# #############################################################################

import wandb
import json
import math
import numpy as np
import matplotlib.pyplot as plt
import time
import os
import gc
import torch
from torch import nn
from lensless.eval.benchmark import benchmark
from lensless.hardware.trainable_mask import TrainableMask
from tqdm import tqdm
from lensless.recon.drunet.network_unet import UNetRes
from lensless.recon.restormer import Restormer
from lensless.utils.io import save_image
from lensless.utils.plot import plot_image
from lensless.utils.dataset import SimulatedDatasetTrainableMask
from lensless.utils.image import rotate_HWC


def double_cnn_max_pool(c_in, c_out, cnn_kernel=3, max_pool=2, padding=1, skip_last_relu=False):
    return nn.Sequential(
        nn.Conv2d(
            in_channels=c_in,
            out_channels=c_out,
            kernel_size=cnn_kernel,
            padding=padding,
            bias=False,
        ),
        nn.BatchNorm2d(c_out),
        nn.ReLU(),
        nn.Conv2d(
            in_channels=c_out,
            out_channels=c_out,
            kernel_size=cnn_kernel,
            padding=padding,
            bias=False,
        ),
        nn.BatchNorm2d(c_out),
        nn.ReLU() if not skip_last_relu else nn.Identity(),
        # don't pass stride=1, otherwise no pooling/downsampling..
        nn.MaxPool2d(kernel_size=max_pool, padding=0) if max_pool else nn.Identity(),
    )


class ResBlock(nn.Module):
    def __init__(self, c_in, c_out, cnn_kernel=3, max_pool=2, padding=1):
        super(ResBlock, self).__init__()
        assert c_in == c_out, "Input and output channels must be the same for residual block."

        # conv layers for residual need to be the same size
        self.double_conv = double_cnn_max_pool(
            c_in, c_in, cnn_kernel=cnn_kernel, max_pool=False, padding=padding, skip_last_relu=True
        )

        # pooling
        self.pooling = nn.Sequential(
            # nn.Conv2d(
            #     in_channels=c_in,
            #     out_channels=c_out,
            #     kernel_size=cnn_kernel,
            #     padding=padding,
            #     bias=False,
            # ),
            # nn.BatchNorm2d(c_out),
            nn.ReLU(),
            nn.MaxPool2d(kernel_size=max_pool, padding=0),
        )
        # self.pooling = nn.MaxPool2d(kernel_size=max_pool, padding=0)

    def forward(self, x):
        return self.pooling(x + self.double_conv(x))


class CompensationBranch(nn.Module):
    """
    Compensation branch for unrolled algorithm, as in "Robust Reconstruction With Deep Learning to Handle Model Mismatch in Lensless Imaging" (2021).
    """

    def __init__(self, nc, cnn_kernel=3, max_pool=2, in_channel=3, residual=True, padding=1):
        """

        Parameters
        ----------
        nc : list
            Number of channels for each layer of the compensation branch.
        cnn_kernel : int, optional
            Kernel size for convolutional layers, by default 3.
        max_pool : int, optional
            Kernel size for max pooling layers, by default 2.
        in_channel : int, optional
            Number of input channels, by default 3 for RGB.
        residual : bool, optional
            Whether to use residual block or simply double conv for intermediate layers, by default True.
        """
        super(CompensationBranch, self).__init__()

        self.n_iter = len(nc)

        # layers along the compensation branch, f^C in paper
        branch_layers = [
            double_cnn_max_pool(
                in_channel,
                nc[0],
                cnn_kernel=cnn_kernel,
                max_pool=max_pool,
                padding=padding,
            )
        ]
        self.branch_layers = nn.ModuleList(
            branch_layers
            + [
                double_cnn_max_pool(
                    # nc[i] * 2,  # due to concatenation with intermediate layer
                    nc[i] + 3,  # due to concatenation with intermediate layer
                    nc[i + 1],
                    cnn_kernel=cnn_kernel,
                    max_pool=max_pool,
                    padding=padding,
                )
                for i in range(self.n_iter - 1)
            ]
        )

        # residual layers for intermediate output, \tilde{f}^C in paper
        # -- not mentinoed in paper, but added more max-pooling for later residual layers, otherwise dimensions don't match
        self.residual_layers = nn.ModuleList(
            [
                ResBlock(
                    in_channel,
                    in_channel,
                    cnn_kernel=cnn_kernel,
                    max_pool=max_pool ** (i + 1),
                    padding=padding,
                )
                if residual
                else double_cnn_max_pool(
                    in_channel,
                    nc[i],
                    cnn_kernel=cnn_kernel,
                    max_pool=max_pool ** (i + 1),
                    padding=padding,
                )
                for i in range(self.n_iter - 1)
            ]
        )

    def forward(self, x, return_NCHW=True):
        """
        Input must be original input and intermediate outputs: (b, s1, s2, ... , s^{K-1}), where K is the number of iterations.

        See p. 1085 of "Robust Reconstruction With Deep Learning to Handle Model Mismatch in Lensless Imaging" (2021) for more details.
        """
        assert len(x) == self.n_iter, "Input must have the same length as the number of iterations."
        n_depth = x[0].shape[-4]
        h_apo_k = self.branch_layers[0](convert_to_NCHW(x[0]))  # h^{'}_k
        for k in range(self.n_iter - 1):  # eq. 18-21
            # \tilde{h}_k
            h_k = torch.cat([h_apo_k, self.residual_layers[k](convert_to_NCHW(x[k + 1]))], axis=1)
            h_apo_k = self.branch_layers[k + 1](h_k)  # h^{'}_k

        if return_NCHW:
            return h_apo_k
        else:
            return convert_to_NDCHW(h_apo_k, n_depth)


# convert from NDHWC to NCHW
def convert_to_NCHW(image):
    image = image.movedim(-1, -3)
    image = image.reshape(-1, *image.shape[-3:])
    return image


# convert back to NDHWC
def convert_to_NDCHW(image, depth):
    image = image.movedim(-3, -1)
    image = image.reshape(-1, depth, *image.shape[-3:])
    return image


def load_drunet(model_path=None, n_channels=3, requires_grad=False):
    """
    Load a pre-trained Drunet model.

    Parameters
    ----------
    model_path : str, optional
        Path to pre-trained model. Download if not provided.
    n_channels : int
        Number of channels in input image.
    requires_grad : bool
        Whether to require gradients for model parameters.

    Returns
    -------
    model : :py:class:`torch.nn.Module`
        Loaded model.
    """

    if model_path is None:
        this_file_path = os.path.dirname(os.path.realpath(__file__))
        model_path = os.path.join(this_file_path, "..", "..", "models", "drunet_color.pth")
        if not os.path.exists(model_path):
            try:
                from torchvision.datasets.utils import download_url
            except ImportError:
                exit()
            msg = "Do you want to download the pretrained DRUNet model (130MB)?"

            # default to yes if no input is given
            valid = input("%s (Y/n) " % msg).lower() != "n"
            output_path = os.path.join(this_file_path, "..", "..", "models")
            if valid:
                url = "https://drive.switch.ch/index.php/s/jTdeMHom025RFRQ/download"
                filename = "drunet_color.pth"
                download_url(url, output_path, filename=filename)

    assert os.path.exists(model_path), f"Model path {model_path} does not exist"

    model = UNetRes(
        in_nc=n_channels + 1,
        out_nc=n_channels,
        nc=[64, 128, 256, 512],
        nb=4,
        act_mode="R",
        downsample_mode="strideconv",
        upsample_mode="convtranspose",
    )
    model.load_state_dict(torch.load(model_path), strict=True)
    model.eval()
    for _, v in model.named_parameters():
        v.requires_grad = requires_grad

    return model


def apply_denoiser(
    model, image, noise_level=10, mode="inference", compensation_output=None, background=None
):
    """
    Apply a pre-trained denoising model with input in the format Channel, Height, Width.
    An additionnal channel is added for the noise level as done in Drunet.

    Parameters
    ----------
    model : :py:class:`torch.nn.Module`
        Drunet compatible model. Its input must consist of 4 channels (RGB + noise level) and output an RGB image both in CHW format.
    image : :py:class:`torch.Tensor`
        Input image.
    noise_level : float or :py:class:`torch.Tensor`
        Noise level in the image within [0, 255].
    background : :py:class:`torch.Tensor`, optional
        If provided, use background as noise channel instead of noise level.
    device : str
        Device to use for computation. Can be "cpu" or "cuda".
    mode : str
        Mode to use for model. Can be "inference" or "train".

    Returns
    -------
    image : :py:class:`torch.Tensor`
        Reconstructed image.
    """
    assert noise_level > 0
    assert noise_level <= 255

    # convert from NDHWC to NCHW
    depth = image.shape[-4]
    image = image.movedim(-1, -3)
    image = image.reshape(-1, *image.shape[-3:])

    # pad image H and W to next multiple of 8
    top = (8 - image.shape[-2] % 8) // 2
    bottom = (8 - image.shape[-2] % 8) - top
    left = (8 - image.shape[-1] % 8) // 2
    right = (8 - image.shape[-1] % 8) - left
    image = torch.nn.functional.pad(image, (left, right, top, bottom), mode="constant", value=0)

    # use background if provided
    # TODO distinguish between integrated background subtraction where it gets passed to the model instead
    # -- check model.background_subtraction ? At the moment, integrated background subtraction isn't supported
    if background is not None:
        # -- pad
        background = background.movedim(-1, -3)
        background = background.reshape(-1, *background.shape[-3:])
        background = torch.nn.functional.pad(
            background, (left, right, top, bottom), mode="constant", value=0
        )

        # -- concatenate as noise channel
        image = torch.cat((image, background), dim=1)

    else:

        # add noise level as extra channel
        if isinstance(noise_level, torch.Tensor):
            noise_level = noise_level / 255.0
        else:
            noise_level = torch.tensor([noise_level / 255.0])
        image = torch.cat(
            (
                image,
                noise_level.repeat(image.shape[0], 1, image.shape[2], image.shape[3]).to(
                    image.device
                ),
            ),
            dim=1,
        )

    # apply model
    if mode == "inference":
        with torch.no_grad():
            image = model(image, compensation_output)
    elif mode == "train":
        image = model(image, compensation_output)
    else:
        raise ValueError("mode must be 'inference' or 'train'")

    # remove padding
    image = image[:, :, top:-bottom, left:-right]
    # convert back to NDHWC
    image = image.movedim(-3, -1)
    image = image.reshape(-1, depth, *image.shape[-3:])

    return image


def get_drunet_function(model, mode="inference"):
    """
    Return a processing function that applies the DruNet model to an image.
    Legacy function to work with pre-trained models, use get_drunet_function_v2 instead.

    Parameters
    ----------
    model : :py:class:`torch.nn.Module`
        DruNet like denoiser model
    device : str
        Device to use for computation. Can be "cpu" or "cuda".
    mode : str
        Mode to use for model. Can be "inference" or "train".
    """

    def process(image, noise_level=10, compensation_output=None, background=None):
        assert compensation_output is None, "Compensation output not supported for legacy models."
        assert background is None, "Background not supported for legacy models."
        x_max = torch.amax(image, dim=(-2, -3), keepdim=True) + 1e-6
        image = apply_denoiser(
            model,
            image,
            noise_level=noise_level,
            mode=mode,
        )
        image = torch.clip(image, min=0.0) * x_max
        return image

    return process


def get_drunet_function_v2(model, mode="inference"):
    """
    Return a processing function that applies the DruNet model to an image.

    Parameters
    ----------
    model : :py:class:`torch.nn.Module`
        DruNet like denoiser model
    mode : str
        Mode to use for model. Can be "inference" or "train".
    """

    def process(image, noise_level=10, compensation_output=None, background=None):
        x_max = torch.amax(image, dim=(-1, -2, -3, -4), keepdim=True) + 1e-6
        image = apply_denoiser(
            model,
            image / x_max,
            noise_level=noise_level,
            mode=mode,
            compensation_output=compensation_output,
            background=background,
        )
        image = torch.clip(image, min=0.0) * x_max.to(image.device)
        return image

    return process


def measure_gradient(model):
    """
    Helper function to measure L2 norm of the gradient of a model.

    Parameters
    ----------
    model : :py:class:`torch.nn.Module`
        Model to measure gradient of.

    Returns
    -------
    Float
        L2 norm of the gradient of the model.
    """
    total_norm = 0.0
    for p in model.parameters():
        param_norm = p.grad.detach().data.norm(2)
        total_norm += param_norm.item() ** 2
    total_norm = total_norm**0.5
    return total_norm


def create_process_network(
    network,
    device="cpu",
    device_ids=None,
    concatenate_compensation=False,
    background_subtraction=False,
    input_background=False,
    # unet parameters
    depth=4,
    nc=None,
    # restormer parameters
    restormer_params=None,
):
    """
    Helper function to create a process network.

    Parameters
    ----------
    network : str
        Name of network to use. Can be "DruNet" or "UnetRes".
    depth : int
        Depth of network.
    device : str
        Device to use for computation. Can be "cpu" or "cuda". Defaults to "cpu".
    concatenate_compensation : int
        Number of channels in last layer of compensation branch.

    Returns
    -------
    :py:class:`torch.nn.Module`
        New process network. Already trained for Drunet.
    """

    if nc is None:
        nc = [64, 128, 256, 512]
    else:
        assert len(nc) == 4

    if network == "DruNet":
        assert (
            concatenate_compensation is False
        ), "DruNet does not support concatenation of compensation branch."
        assert background_subtraction is False, "DruNet does not support background subtraction."
        from lensless.recon.utils import load_drunet

        process = load_drunet(requires_grad=True)
        process_name = "DruNet"
    elif network == "UnetRes":
        n_channels = 3
        process = UNetRes(
            in_nc=n_channels * 2
            if input_background
            else n_channels + 1,  # extra channel(s) for noise level(s)
            out_nc=n_channels,
            nc=nc,
            nb=depth,
            act_mode="R",
            downsample_mode="strideconv",
            upsample_mode="convtranspose",
            concatenate_compensation=concatenate_compensation,
            input_background=input_background or background_subtraction,
            background_subtraction=background_subtraction,
        )
        process_name = "UnetRes_d" + str(depth)

    elif network == "Restormer":
        assert restormer_params is not None
        process = Restormer(
            inp_channels=3,
            out_channels=3,
            dim=restormer_params["dim"],
            num_blocks=restormer_params["num_blocks"],
            num_refinement_blocks=restormer_params["num_refinement_blocks"],
            heads=restormer_params["heads"],
            ffn_expansion_factor=restormer_params["ffn_expansion_factor"],
            bias=False,
            LayerNorm_type="BiasFree",  # "WithBias"
            dual_pixel_task=False,
        )

        ## -- pretrained
        # process = Restormer(
        #     inp_channels=3,
        #     out_channels=3,
        #     dim=48,
        #     num_blocks=[4, 6, 6, 8],
        #     num_refinement_blocks=4,
        #     heads=[1, 2, 4, 8],
        #     ffn_expansion_factor=2.66,
        #     bias=False,
        #     LayerNorm_type="BiasFree",  # "WithBias"
        #     dual_pixel_task=False,
        # )
        # # TODO download path from https://drive.google.com/drive/folders/1Qwsjyny54RZWa7zC4Apg7exixLBo4uF0
        # weights_path = "/root/LenslessPiCam/notebook/real_denoising.pth"
        # checkpoint = torch.load(weights_path)
        # process.load_state_dict(checkpoint['params'])

        process_name = "Restormer"

    else:
        process = None
        process_name = None

    if process is not None:
        if device_ids is not None:
            process = torch.nn.DataParallel(process, device_ids=device_ids)
        process = process.to(device)

    return (process, process_name)


class Trainer:
    def __init__(
        self,
        recon,
        train_dataset,
        test_dataset,
        test_size=0.15,
        mask=None,
        batch_size=4,
        eval_batch_size=10,
        loss="l2",
        lpips=None,
        l1_mask=None,
        optimizer=None,
        skip_NAN=False,
        algorithm_name="Unknown",
        metric_for_best_model=None,
        save_every=None,
        gamma=None,
        logger=None,
        crop=None,
        clip_grad=1.0,
        unrolled_output_factor=False,
        random_rotate=False,
        random_shift=False,
        pre_proc_aux=False,
        extra_eval_sets=None,
        use_wandb=False,
        # for adding components during training
        pre_process=None,
        pre_process_delay=None,
        pre_process_freeze=None,
        pre_process_unfreeze=None,
        post_process=None,
        post_process_delay=None,
        post_process_freeze=None,
        post_process_unfreeze=None,
        n_epoch=None,
    ):
        """
        Class to train a reconstruction algorithm. Inspired by Trainer from `HuggingFace <https://huggingface.co/docs/transformers/main_classes/trainer>`__.

        The train and test metrics at the end of each epoch can be found in ``self.metrics``,
        with "LOSS" being the train loss. The test loss can be found in "MSE" (if loss is "l2") or
        "MAE" (if loss is "l1"). If ``lpips`` is not None, the LPIPS loss is also added
        to the train loss, such that the test loss can be computed as "MSE" + ``lpips`` * "LPIPS_Vgg"
        (or "MAE" + ``lpips`` * "LPIPS_Vgg").

        Parameters
        ----------
        recon : :py:class:`lensless.TrainableReconstructionAlgorithm`
            Reconstruction algorithm to train.
        train_dataset : :py:class:`torch.utils.data.Dataset`
            Dataset to use for training.
        test_dataset : :py:class:`torch.utils.data.Dataset`
            Dataset to use for testing.
        test_size : float, optional
            If test_dataset is None, fraction of the train dataset to use for testing, by default 0.15.
        mask : TrainableMask, optional
            Trainable mask to use for training. If none, training with fix psf, by default None.
        batch_size : int, optional
            Batch size to use for training, by default 4.
        eval_batch_size : int, optional
            Batch size to use for evaluation, by default 10.
        loss : str, optional
            Loss function to use for training "l1" or "l2", by default "l2".
        lpips : float, optional
            the weight of the lpips(VGG) in the total loss. If None ignore. By default None.
        l1_mask : float, optional
            the weight of the l1 norm of the mask in the total loss. If None ignore. By default None.
        optimizer : dict
            Optimizer configuration.
        skip_NAN : bool, optional
            Whether to skip update if any gradiant are NAN (True) or to throw an error(False), by default False
        algorithm_name : str, optional
            Algorithm name for logging, by default "Unknown".
        metric_for_best_model : str, optional
            Metric to use for saving the best model. If None, will default to evaluation loss. Default is None.
        save_every : int, optional
            Save model every ``save_every`` epochs. If None, just save best model.
        gamma : float, optional
            Gamma correction to apply to PSFs when plotting. If None, no gamma correction is applied. Default is None.
        logger : :py:class:`logging.Logger`, optional
            Logger to use for logging. If None, just print to terminal. Default is None.
        crop : dict, optional
            Crop to apply to images before computing loss (by applying a mask). If None, no crop is applied. Default is None.
        unrolled_output_factor : float, optional
            How much of the unrolled loss to add to the total loss. If False, no unrolled loss is added. Default is False. Only applicable if a post-processor is used.
        pre_process : :py:class:`torch.nn.Module`, optional
            Pre process component to add during training. Default is None.
        pre_process_delay : int, optional
            Epoch at which to add pre process component. Default is None.
        pre_process_freeze : int, optional
            Epoch at which to freeze pre process component. Default is None.
        pre_process_unfreeze : int, optional
            Epoch at which to unfreeze pre process component. Default is None.
        post_process : :py:class:`torch.nn.Module`, optional
            Post process component to add during training. Default is None.
        post_process_delay : int, optional
            Epoch at which to add post process component. Default is None.
        post_process_freeze : int, optional
            Epoch at which to freeze post process component. Default is None.
        post_process_unfreeze : int, optional
            Epoch at which to unfreeze post process component. Default is None.

        """
        # global print

        self.use_wandb = use_wandb

        self.device = recon._psf.device
        self.logger = logger
        if self.logger is not None:
            self.print = self.logger.info
        else:
            self.print = print
        self.recon = recon

        self.pre_process = pre_process
        self.pre_process_delay = pre_process_delay
        self.pre_process_freeze = pre_process_freeze
        self.pre_process_unfreeze = pre_process_unfreeze
        self.pre_process_delay = pre_process_delay
        if pre_process_delay is not None:
            assert pre_process is not None

        self.post_process = post_process
        self.post_process_delay = post_process_delay
        self.post_process_freeze = post_process_freeze
        self.post_process_unfreeze = post_process_unfreeze
        self.post_process_delay = post_process_delay
        if post_process_delay is not None:
            assert post_process is not None

        assert train_dataset is not None
        if test_dataset is None:
            assert test_size < 1.0 and test_size > 0.0
            # split train dataset
            train_size = int((1 - test_size) * len(train_dataset))
            test_size = len(train_dataset) - train_size
            train_dataset, test_dataset = torch.utils.data.random_split(
                train_dataset, [train_size, test_size]
            )
            self.print(f"Train size : {train_size}, Test size : {test_size}")

        self.train_dataset = train_dataset
        self.train_dataloader = torch.utils.data.DataLoader(
            dataset=train_dataset,
            batch_size=batch_size,
            shuffle=True,
            pin_memory=(self.device != "cpu"),
        )
        self.test_dataset = test_dataset
        self.extra_eval_sets = extra_eval_sets  # additional datasets to evaluate on
        self.lpips = lpips
        self.skip_NAN = skip_NAN
        self.eval_batch_size = eval_batch_size
        self.train_multimask = False
        if hasattr(train_dataset, "multimask"):
            self.train_multimask = train_dataset.multimask
        self.train_random_flip = train_dataset.random_flip
        self.random_rotate = random_rotate
        self.random_shift = random_shift
        if hasattr(train_dataset, "measured_bg"):
            self.background = train_dataset.measured_bg
        else:
            self.background = False
        if self.random_shift:
            raise NotImplementedError("Random shift not implemented yet.")

        # check if Subset and if simulating dataset
        self.simulated_dataset_trainable_mask = False
        if isinstance(self.test_dataset, SimulatedDatasetTrainableMask):
            # assuming the case for both training and testing
            self.simulated_dataset_trainable_mask = True

        self.mask = mask
        self.gamma = gamma
        if mask is not None:
            assert isinstance(mask, TrainableMask)
            self.use_mask = True
        else:
            self.use_mask = False

        if self.use_mask:
            # save original PSF
            psf_np = self.mask.get_psf().detach().cpu().numpy()
            np.save("psf_original.npy", psf_np)
            n_psf = psf_np.shape[0]
            _, ax = plt.subplots(n_psf, 1)
            for i in range(n_psf):
                psf_np_i = psf_np[i].squeeze()
                fp = f"psf_original_{i}.png" if n_psf > 1 else "psf_original.png"
                save_image(psf_np_i, fp)
                plot_image(psf_np_i, gamma=self.gamma, ax=ax[i] if n_psf > 1 else ax)
                if n_psf > 1:
                    ax[i].axis("off")
                else:
                    ax.axis("off")

                if self.use_wandb and save_every is not None:
                    log_key = f"psf_{i}" if n_psf > 1 else "psf"
                    wandb.log({log_key: wandb.Image(fp)}, step=0)

            # save plot
            fp = "psf_original_PLOT.png"
            plt.savefig(fp)

        self.l1_mask = l1_mask

        # loss
        if loss == "l2":
            self.Loss = torch.nn.MSELoss()
        elif loss == "l1":
            self.Loss = torch.nn.L1Loss()
        else:
            raise ValueError(f"Unsuported loss : {loss}")

        # -- Lpips loss
        if lpips:
            try:
                import lpips

                self.Loss_lpips = lpips.LPIPS(net="vgg").to(self.device)
            except ImportError:
                return ImportError(
                    "lpips package is need for LPIPS loss. Install using : pip install lpips"
                )

        self.crop = crop

        # -- adding unrolled loss
        self.unrolled_output_factor = unrolled_output_factor
        if self.unrolled_output_factor:
            assert self.unrolled_output_factor > 0
            assert self.post_process is not None
            assert self.post_process_delay is None
            assert self.post_process_unfreeze is None
            assert self.post_process_freeze is None

        # -- adding pre-processed output to loss
        self.pre_proc_aux = pre_proc_aux
        if self.pre_proc_aux:
            assert self.pre_process is not None
            assert self.pre_process_delay is None
            assert self.pre_process_unfreeze is None
            assert self.pre_process_freeze is None

        # optimizer
        self.clip_grad_norm = clip_grad
        self.optimizer_config = optimizer
        self.n_epoch = n_epoch
        self.lr_step_epoch = optimizer.lr_step_epoch
        self.set_optimizer()

        # metrics
        self.metrics = {
            "LOSS": [],  # train loss
            "LOSS_TEST": [],  # test loss
            "MSE": [],
            "MAE": [],
            "LPIPS_Vgg": [],
            "LPIPS_Alex": [],
            "PSNR": [],
            "SSIM": [],
            "ReconstructionError": [],
            "n_iter": self.recon._n_iter,
            "algorithm": algorithm_name,
            "metric_for_best_model": metric_for_best_model,
            "best_epoch": 0,
            "best_eval_score": 0
            if metric_for_best_model == "PSNR" or metric_for_best_model == "SSIM"
            else np.inf,
        }
        if self.unrolled_output_factor:
            # -- add unrolled metrics
            for key in ["MSE", "MAE", "LPIPS_Vgg", "LPIPS_Alex", "PSNR", "SSIM"]:
                self.metrics[key + "_unrolled"] = []
        if self.pre_proc_aux:
            self.metrics[
                "ReconstructionError_PreProc"
            ] = []  # reconstruction error of ||pre_proc(y) - A * camera_inversion(y)||
        if metric_for_best_model is not None:
            assert metric_for_best_model in self.metrics.keys()
        if extra_eval_sets is not None:
            for key in extra_eval_sets:
                self.metrics[key] = dict()
        self.save_every = save_every

        # Backward hook that detect NAN in the gradient and print the layer weights
        if not self.skip_NAN:

            def detect_nan(grad):
                if torch.isnan(grad).any():
                    if self.logger:
                        self.logger.info(grad)
                    else:
                        print(grad, flush=True)
                    for name, param in recon.named_parameters():
                        if param.requires_grad:
                            self.print(name, param)
                    raise ValueError("Gradient is NaN")
                return grad

            for param in recon.parameters():
                if param.requires_grad:
                    param.register_hook(detect_nan)
                    if param.requires_grad:
                        param.register_hook(detect_nan)

    def set_optimizer(self, last_epoch=-1):

        if self.optimizer_config.type == "AdamW":
            print("USING ADAMW")
            self.optimizer = torch.optim.AdamW(
                [
                    {"params": [p for p in self.recon.parameters() if p.dim() > 1]},
                    {
                        "params": [p for p in self.recon.parameters() if p.dim() <= 1],
                        "weight_decay": 0,
                    },  # no weight decay on bias terms
                ],
                lr=self.optimizer_config.lr,
                weight_decay=0.01,
            )
        else:
            print(f"USING {self.optimizer_config.type}")
            parameters = [{"params": self.recon.parameters()}]
            self.optimizer = getattr(torch.optim, self.optimizer_config.type)(
                parameters, lr=self.optimizer_config.lr
            )

        # Scheduler
        if self.optimizer_config.slow_start:

            def learning_rate_function(epoch):
                if epoch == 0:
                    return self.optimizer_config.slow_start
                elif epoch == 1:
                    return math.sqrt(self.optimizer_config.slow_start)
                else:
                    return 1

            self.scheduler = torch.optim.lr_scheduler.LambdaLR(
                self.optimizer, lr_lambda=learning_rate_function, last_epoch=last_epoch
            )

        elif self.optimizer_config.final_lr:

            assert self.optimizer_config.final_lr < self.optimizer_config.lr
            assert self.n_epoch is not None

            # # linear decay
            # def learning_rate_function(epoch):
            #     slope = (start / final - 1) / (n_epoch)
            #     return 1 / (1 + slope * epoch)

            # exponential decay
            def learning_rate_function(epoch):
                final_decay = self.optimizer_config.final_lr / self.optimizer_config.lr
                final_decay = final_decay ** (1 / (self.n_epoch - 1))
                return final_decay**epoch

            self.scheduler = torch.optim.lr_scheduler.LambdaLR(
                self.optimizer, lr_lambda=learning_rate_function, last_epoch=last_epoch
            )

        elif self.optimizer_config.exp_decay:

            def learning_rate_function(epoch):
                return self.optimizer_config.exp_decay**epoch

            self.scheduler = torch.optim.lr_scheduler.LambdaLR(
                self.optimizer, lr_lambda=learning_rate_function, last_epoch=last_epoch
            )

        elif self.optimizer_config.cosine_decay_warmup:

            if self.lr_step_epoch:
                total_iterations = self.n_epoch
            else:
                total_iterations = len(self.train_dataloader) * self.n_epoch
            warmup_steps = int(0.05 * total_iterations)

            def cosine_decay_with_warmup(step, warmup_steps, total_steps):
                if step < warmup_steps:
                    return step / warmup_steps
                progress = (step - warmup_steps) / (total_steps - warmup_steps)
                return 0.5 * (1 + math.cos(math.pi * progress))

            self.scheduler = torch.optim.lr_scheduler.LambdaLR(
                self.optimizer,
                lr_lambda=lambda step: cosine_decay_with_warmup(
                    step, warmup_steps, total_iterations
                ),
            )

        elif self.optimizer_config.step:

            self.scheduler = torch.optim.lr_scheduler.StepLR(
                self.optimizer,
                step_size=self.optimizer_config.step,
                gamma=self.optimizer_config.gamma,
                last_epoch=last_epoch,
                verbose=True,
            )

        else:

            def learning_rate_function(epoch):
                return 1

            self.scheduler = torch.optim.lr_scheduler.LambdaLR(
                self.optimizer, lr_lambda=learning_rate_function, last_epoch=last_epoch
            )

    def train_epoch(self, data_loader):
        """
        Train for one epoch.

        Parameters
        ----------
        data_loader : :py:class:`torch.utils.data.DataLoader`
            Data loader to use for training.

        Returns
        -------
        float
            Mean loss of the epoch.
        """
        mean_loss = 0.0
        i = 1.0
        pbar = tqdm(data_loader)
        self.recon.train()
        for batch in pbar:

            # get batch
            flip_lr = None
            flip_ud = None
            background = None
            X = batch[0].to(self.device)
            y = batch[1].to(self.device)
            if self.background:
                background = batch[-1].to(self.device)
            if self.train_random_flip or self.train_multimask:
                psfs = batch[2].to(self.device)
            else:
                psfs = None
            if self.train_random_flip:
                flip_lr = batch[3]
                flip_ud = batch[4]

            random_rotate = False
            if self.random_rotate:
                random_rotate = np.random.uniform(-self.random_rotate, self.random_rotate)
                X = rotate_HWC(X, random_rotate)
                y = rotate_HWC(y, random_rotate)
                if psfs is None:
                    psf_single = self.recon._psf
                    psf_single = rotate_HWC(psf_single, random_rotate)
                    self.recon._set_psf(psf_single.to(self.device))
                else:
                    psfs = rotate_HWC(psfs, random_rotate)

            # update psf according to mask
            if self.use_mask:
                self.recon._set_psf(self.mask.get_psf().to(self.device))

            # forward pass
            # torch.autograd.set_detect_anomaly(True)    # for debugging
            y_pred = self.recon.forward(batch=X, psfs=psfs, background=background)
            if self.unrolled_output_factor or self.pre_proc_aux:
                y_pred, camera_inv_out, pre_proc_out = y_pred[0], y_pred[1], y_pred[2]

            # normalizing each output
            eps = 1e-12
            y_pred_max = torch.amax(y_pred, dim=(-1, -2, -3), keepdim=True) + eps
            y_pred = y_pred / y_pred_max

            # normalizing y
            y_max = torch.amax(y, dim=(-1, -2, -3), keepdim=True) + eps
            y = y / y_max

            # convert to CHW for loss and remove depth
            y_pred_crop = y_pred.reshape(-1, *y_pred.shape[-3:]).movedim(-1, -3)
            y = y.reshape(-1, *y.shape[-3:]).movedim(-1, -3)

            # extraction region of interest for loss
            if hasattr(self.train_dataset, "alignment"):
                if self.train_dataset.alignment is not None:
                    y_pred_crop = self.train_dataset.extract_roi(
                        y_pred_crop,
                        axis=(-2, -1),
                        flip_lr=flip_lr,
                        flip_ud=flip_ud,
                        rotate_aug=random_rotate,
                    )
                else:
                    y_pred_crop, y = self.train_dataset.extract_roi(
                        y_pred_crop,
                        axis=(-2, -1),
                        lensed=y,
                        flip_lr=flip_lr,
                        flip_ud=flip_ud,
                        rotate_aug=random_rotate,
                    )

            elif self.crop is not None:
                assert flip_lr is None and flip_ud is None
                y_pred_crop = y_pred_crop[
                    ...,
                    self.crop["vertical"][0] : self.crop["vertical"][1],
                    self.crop["horizontal"][0] : self.crop["horizontal"][1],
                ]
                y = y[
                    ...,
                    self.crop["vertical"][0] : self.crop["vertical"][1],
                    self.crop["horizontal"][0] : self.crop["horizontal"][1],
                ]

            loss_v = self.Loss(y_pred_crop, y)

            # add LPIPS loss
            if self.lpips:

                if y_pred_crop.shape[1] == 1:
                    # if only one channel, repeat for LPIPS
                    y_pred_crop = y_pred_crop.repeat(1, 3, 1, 1)
                    y = y.repeat(1, 3, 1, 1)

                # value for LPIPS needs to be in range [-1, 1]
                loss_v = loss_v + self.lpips * torch.mean(
                    self.Loss_lpips(2 * y_pred_crop - 1, 2 * y - 1)
                )
            if self.use_mask and self.l1_mask:
                for p in self.mask.parameters():
                    if p.requires_grad:
                        loss_v = loss_v + self.l1_mask * torch.mean(torch.abs(p))

            if self.unrolled_output_factor:
                # -- normalize
                unrolled_out_max = torch.amax(camera_inv_out, dim=(-1, -2, -3), keepdim=True) + eps
                camera_inv_out_norm = camera_inv_out / unrolled_out_max

                # -- convert to CHW for loss and remove depth
                camera_inv_out_norm = camera_inv_out_norm.reshape(
                    -1, *camera_inv_out.shape[-3:]
                ).movedim(-1, -3)

                # -- extraction region of interest for loss
                if hasattr(self.train_dataset, "alignment"):
                    if self.train_dataset.alignment is not None:
                        camera_inv_out_norm = self.train_dataset.extract_roi(
                            camera_inv_out_norm, axis=(-2, -1)
                        )
                    else:
                        camera_inv_out_norm = self.train_dataset.extract_roi(
                            camera_inv_out_norm,
                            axis=(-2, -1),
                            # y=y   # lensed already extracted before
                        )
                    assert np.all(y.shape == camera_inv_out_norm.shape)
                elif self.crop is not None:
                    camera_inv_out_norm = camera_inv_out_norm[
                        ...,
                        self.crop["vertical"][0] : self.crop["vertical"][1],
                        self.crop["horizontal"][0] : self.crop["horizontal"][1],
                    ]

                # -- compute unrolled output loss
                loss_unrolled = self.Loss(camera_inv_out_norm, y)

                # -- add LPIPS loss
                if self.lpips:
                    if camera_inv_out_norm.shape[1] == 1:
                        # if only one channel, repeat for LPIPS
                        camera_inv_out_norm = camera_inv_out_norm.repeat(1, 3, 1, 1)

                    # value for LPIPS needs to be in range [-1, 1]
                    loss_unrolled = loss_unrolled + self.lpips * torch.mean(
                        self.Loss_lpips(2 * camera_inv_out_norm - 1, 2 * y - 1)
                    )

                # -- add unrolled loss to total loss
                loss_v = loss_v + self.unrolled_output_factor * loss_unrolled

            if self.pre_proc_aux:
                # -- normalize
                unrolled_out_max = torch.amax(camera_inv_out, dim=(-1, -2, -3), keepdim=True) + eps
                camera_inv_out_norm = camera_inv_out / unrolled_out_max

                err = torch.mean(
                    self.recon.reconstruction_error(
                        prediction=camera_inv_out_norm,
                        # prediction=y_pred,
                        lensless=pre_proc_out,
                    )
                )
                loss_v = loss_v + self.pre_proc_aux * err

            # backward pass
            loss_v.backward()

            # check mask parameters are learning
            if self.use_mask:
                for p in self.mask.parameters():
                    assert p.grad is not None

            if self.clip_grad_norm is not None:
                if self.use_mask:
                    torch.nn.utils.clip_grad_norm_(self.mask.parameters(), self.clip_grad_norm)
                torch.nn.utils.clip_grad_norm_(self.recon.parameters(), self.clip_grad_norm)

            # if any gradient is NaN, skip training step
            if self.skip_NAN:
                recon_is_NAN = False
                mask_is_NAN = False
                for param in self.recon.parameters():
                    if param.grad is not None and torch.isnan(param.grad).any():
                        recon_is_NAN = True
                        break
                if self.use_mask:
                    for param in self.mask.parameters():
                        if param.grad is not None and torch.isnan(param.grad).any():
                            mask_is_NAN = True
                            break
                if recon_is_NAN or mask_is_NAN:
                    if recon_is_NAN:
                        self.print(
                            "NAN detected in reconstruction gradient, skipping training step"
                        )
                    if mask_is_NAN:
                        self.print("NAN detected in mask gradient, skipping training step")
                    i += 1
                    continue

            self.optimizer.step()
            if not self.lr_step_epoch:
                self.scheduler.step()
            self.optimizer.zero_grad(set_to_none=True)

            # update mask
            if self.use_mask:
                self.mask.update_mask()
                if self.simulated_dataset_trainable_mask:
                    self.train_dataloader.dataset.set_psf()

            mean_loss += (loss_v.item() - mean_loss) * (1 / i)
            pbar.set_description(f"loss : {mean_loss}")
            i += 1

        self.print(f"loss : {mean_loss}")

        return mean_loss

    def evaluate(self, mean_loss, epoch, disp=None):
        """
        Evaluate the reconstruction algorithm on the test dataset.

        Parameters
        ----------
        mean_loss : float
            Mean loss of the last epoch.
        disp : list of int, optional
            Test set examples to visualize at the end of each epoch, by default None.
        """
        if self.test_dataset is None:
            return

        if self.use_mask and self.simulated_dataset_trainable_mask:
            with torch.no_grad():
                self.test_dataset.set_psf()

        output_dir = None
        if disp is not None:
            output_dir = os.path.join("eval_recon")
            if not os.path.exists(output_dir):
                os.mkdir(output_dir)
            output_dir = os.path.join(output_dir, str(epoch))

        # benchmarking
        self.recon.eval()
        current_metrics = benchmark(
            self.recon,
            self.test_dataset,
            batchsize=self.eval_batch_size,
            save_idx=disp,
            output_dir=output_dir,
            crop=self.crop,
            unrolled_output_factor=self.unrolled_output_factor,
            pre_process_aux=self.pre_proc_aux,
            use_wandb=self.use_wandb,
            epoch=epoch,
        )

        # update metrics with current metrics
        self.metrics["LOSS"].append(mean_loss)
        if self.use_wandb:
            wandb.log({"LOSS": mean_loss}, step=epoch)
        for key in current_metrics:
            self.metrics[key].append(current_metrics[key])

        # check best metric
        if self.metrics["metric_for_best_model"] is None:
            eval_loss = current_metrics["MSE"]
            if self.lpips is not None:
                eval_loss += self.lpips * current_metrics["LPIPS_Vgg"]
            if self.use_mask and self.l1_mask:
                with torch.no_grad():
                    for p in self.mask.parameters():
                        if p.requires_grad:
                            eval_loss += self.l1_mask * np.mean(np.abs(p.cpu().detach().numpy()))
            if self.unrolled_output_factor:
                unrolled_loss = current_metrics["MSE_unrolled"]
                if self.lpips is not None:
                    unrolled_loss += self.lpips * current_metrics["LPIPS_Vgg_unrolled"]
                eval_loss += self.unrolled_output_factor * unrolled_loss
            if self.pre_proc_aux:
                eval_loss += self.pre_proc_aux * current_metrics["ReconstructionError_PreProc"]
        else:
            eval_loss = current_metrics[self.metrics["metric_for_best_model"]]

        self.metrics["LOSS_TEST"].append(eval_loss)
        if self.use_wandb:
            wandb.log({"LOSS_TEST": eval_loss}, step=epoch)

        # add extra evaluation sets
        extra_metrics_epoch = {}
        if self.extra_eval_sets is not None:
            for eval_set in self.extra_eval_sets:

                # create output directory
                output_dir = None
                if disp is not None:
                    output_dir = os.path.join("eval_recon")
                    if not os.path.exists(output_dir):
                        os.mkdir(output_dir)
                    output_dir = os.path.join(output_dir, str(epoch) + f"_{eval_set}")

                if hasattr(self.extra_eval_sets[eval_set], "multimask"):
                    if not self.extra_eval_sets[eval_set].multimask:
                        # need to set correct PSF for evaluation
                        # TODO cleaner way to set PSF?
                        self.recon._set_psf(self.extra_eval_sets[eval_set].psf.to(self.device))

                # benchmarking
                extra_metrics = benchmark(
                    self.recon,
                    self.extra_eval_sets[eval_set],
                    batchsize=self.eval_batch_size,
                    save_idx=disp,
                    output_dir=output_dir,
                    crop=self.crop,
                    unrolled_output_factor=self.unrolled_output_factor,
                    use_wandb=self.use_wandb,
                    label=eval_set,
                    epoch=epoch,
                )

                # add metrics to dictionary
                for key in extra_metrics:
                    if key not in self.metrics[eval_set]:
                        self.metrics[eval_set][key] = [extra_metrics[key]]
                    else:
                        self.metrics[eval_set][key].append(extra_metrics[key])
                    extra_metrics_epoch[f"{eval_set}_{key}"] = extra_metrics[key]

            # set back PSF to original in case changed
            # TODO: cleaner way?
            if not self.train_multimask:
                self.recon._set_psf(self.train_dataset.psf.to(self.device))

        # log metrics to wandb
        if self.use_wandb:
            wandb.log(current_metrics, step=epoch)
            if self.extra_eval_sets is not None:
                wandb.log(extra_metrics_epoch, step=epoch)

        return eval_loss

    def on_epoch_end(self, mean_loss, save_pt, epoch, disp=None):
        """
        Called at the end of each epoch.

        Parameters
        ----------
        mean_loss : float
            Mean loss of the last epoch.
        save_pt : str
            Path to save metrics dictionary to. If None, no logging of metrics.
        epoch : int
            Current epoch.
        disp : list of int, optional
            Test set examples to visualize at the end of each epoch, by default None.
        """
        if save_pt is None:
            # Use current directory
            save_pt = os.getcwd()

        # save model
        epoch_eval_metric = self.evaluate(mean_loss, epoch, disp=disp)
        new_best = False
        if (
            self.metrics["metric_for_best_model"] == "PSNR"
            or self.metrics["metric_for_best_model"] == "SSIM"
        ):
            if epoch_eval_metric > self.metrics["best_eval_score"]:
                self.metrics["best_eval_score"] = epoch_eval_metric
                new_best = True
        else:
            if epoch_eval_metric < self.metrics["best_eval_score"]:
                self.metrics["best_eval_score"] = epoch_eval_metric
                new_best = True

        if new_best:
            self.metrics["best_epoch"] = epoch
            self.save(path=save_pt, include_optimizer=False, epoch="BEST")

        if self.save_every is not None and epoch % self.save_every == 0:
            self.save(path=save_pt, include_optimizer=False, epoch=epoch)

        # save dictionary metrics to file with json
        with open(os.path.join(save_pt, "metrics.json"), "w") as f:
            json.dump(self.metrics, f, indent=4)

    def train(self, n_epoch=1, save_pt=None, disp=None):
        """
        Train the reconstruction algorithm.

        Parameters
        ----------
        n_epoch : int, optional
            Number of epochs to train for, by default 1
        save_pt : str, optional
            Path to save metrics dictionary to. If None, use current directory, by default None
        disp : list of int, optional
            test set examples to visualize at the end of each epoch, by default None.
        """

        start_time = time.time()

        self.evaluate(mean_loss=1, epoch=0, disp=disp)
        for epoch in range(n_epoch):

            # add extra components (if specified)
            changing_n_param = False
            if epoch == self.pre_process_delay:
                self.print("Adding pre process component")
                self.recon.set_pre_process(self.pre_process)
                changing_n_param = True
            if epoch == self.post_process_delay:
                self.print("Adding post process component")
                self.recon.set_post_process(self.post_process)
                changing_n_param = True
            if epoch == self.pre_process_freeze:
                self.print("Freezing pre process")
                self.recon.freeze_pre_process()
                changing_n_param = True
            if epoch == self.post_process_freeze:
                self.print("Freezing post process")
                self.recon.freeze_post_process()
                changing_n_param = True
            if epoch == self.pre_process_unfreeze:
                self.print("Unfreezing pre process")
                self.recon.unfreeze_pre_process()
                changing_n_param = True
            if epoch == self.post_process_unfreeze:
                self.print("Unfreezing post process")
                self.recon.unfreeze_post_process()
                changing_n_param = True

            # count number of parameters with requires_grad = True
            if changing_n_param:
                n_param = sum(p.numel() for p in self.recon.parameters() if p.requires_grad)
                if self.mask is not None:
                    n_param += sum(p.numel() for p in self.mask.parameters() if p.requires_grad)
                self.print(f"Training {n_param} parameters")

            self.print(f"Epoch {epoch} with learning rate {self.scheduler.get_last_lr()}")
            mean_loss = self.train_epoch(self.train_dataloader)
            # offset because of evaluate before loop
            self.on_epoch_end(mean_loss, save_pt, epoch + 1, disp=disp)
            if self.lr_step_epoch:
                self.scheduler.step()

            torch.cuda.empty_cache()
            gc.collect()

        self.print(f"Train time [hour] : {(time.time() - start_time) / 3600} h")

    def save(self, epoch, path="recon", include_optimizer=False):
        # create directory if it does not exist
        if not os.path.exists(path):
            os.makedirs(path)

        # save mask parameters
        if self.use_mask:

            for name, param in self.mask.named_parameters():
                # save as numpy array
                if param.requires_grad:
                    np.save(
                        os.path.join(path, f"mask{name}_epoch{epoch}.npy"),
                        param.cpu().detach().numpy(),
                    )

            torch.save(
                self.mask._optimizer.state_dict(), os.path.join(path, f"mask_optim_epoch{epoch}.pt")
            )

            psf_np = self.mask.get_psf().detach().cpu().numpy()
            np.save(os.path.join(path, f"psf_epoch{epoch}.npy"), psf_np)
            n_psf = psf_np.shape[0]
            _, ax = plt.subplots(n_psf, 1)
            for i in range(n_psf):
                psf_np_i = psf_np[i].squeeze()
                fp = os.path.join(
                    path, f"psf_epoch{epoch}_{i}.png" if n_psf > 1 else f"psf_epoch{epoch}.png"
                )
                save_image(psf_np_i, fp)
                plot_image(psf_np_i, gamma=self.gamma, ax=ax[i] if n_psf > 1 else ax)
                if n_psf > 1:
                    ax[i].axis("off")
                else:
                    ax.axis("off")

                if self.use_wandb and epoch != "BEST":
                    log_key = f"psf_{i}" if n_psf > 1 else "psf"
                    wandb.log({log_key: wandb.Image(fp)}, step=epoch)

            fp = os.path.join(path, f"psf_epoch{epoch}_PLOT.png")
            plt.savefig(fp)

            if epoch == "BEST":
                # save difference with original PSF
                psf_original = np.load("psf_original.npy")
                diff = psf_np - psf_original
                np.save(os.path.join(path, "psf_epochBEST_diff.npy"), diff)
                diff_abs = np.abs(diff)

                _, ax = plt.subplots(n_psf, 1)
                for i in range(n_psf):
                    diff_abs_i = diff_abs[i].squeeze()
                    save_image(
                        diff_abs_i,
                        os.path.join(path, f"psf_epochBEST_diffabs_{i}.png")
                        if n_psf > 1
                        else os.path.join(path, "psf_epochBEST_diffabs.png"),
                    )
                    if n_psf > 1:
                        ax[i].imshow(diff_abs_i, cmap="gray" if diff_abs.ndim == 2 else None)
                        ax[i].axis("off")
                        ax[0].set_title("Absolute difference with original PSF")
                    else:
                        ax.imshow(diff_abs_i, cmap="gray" if diff_abs.ndim == 2 else None)
                        ax.axis("off")
                        ax.set_title("Absolute difference with original PSF")

                plt.savefig(os.path.join(path, "psf_epochBEST_diffabs_PLOT.png"))

        # save optimizer
        if include_optimizer:
            torch.save(self.optimizer.state_dict(), os.path.join(path, f"optim_epoch{epoch}.pt"))

        # save recon
        torch.save(self.recon.state_dict(), os.path.join(path, f"recon_epoch{epoch}"))



# #############################################################################
# apgd.py
# =================
# Authors :
# Eric BEZZAM [ebezzam@gmail.com]
# Julien SAHLI [julien.sahli@epfl.ch]
# #############################################################################


from lensless.recon.recon import ReconstructionAlgorithm
import inspect
import numpy as np
from typing import Optional
from lensless.utils.image import resize
from lensless.recon.rfft_convolve import RealFFTConvolve2D as Convolver
import cv2

import pycsou.abc as pyca
import pycsou.operator.func as func
import pycsou.opt.solver as solver
import pycsou.opt.stop as stop
import pycsou.runtime as pycrt
import pycsou.util as pycu
import pycsou.util.ptype as pyct
import pycsou.operator.linop as pycl


class APGDPriors:
    """
    Priors (compatible with Pycsou) for APGD.

    """

    L2 = "l2"
    NONNEG = "nonneg"
    L1 = "l1"

    @staticmethod
    def all_values():
        vals = []
        for i in inspect.getmembers(APGDPriors):
            # remove private and protected functions, and this function
            if not i[0].startswith("_") and not callable(i[1]):
                vals.append(i[1])
        return vals


class RealFFTConvolve2D(pyca.LinOp):
    def __init__(
        self, filter: pyct.NDArray, dtype: Optional[type] = None, norm: str = "ortho", **kwargs
    ):
        """
        Linear operator that performs convolution in Fourier domain, and assumes
        real-valued signals.

        Parameters
        ----------
        filter :py:class:`~numpy.ndarray`
            2D filter to use. Must be of shape (height, width, channels) even if
            only one channel.
        dtype : float32 or float64
            Data type to use for optimization.
        norm : str
            Normalization to use for convolution. See :py:class:`~lensless.rfft_convolve.RealFFTConvolve2D`
        """
        assert len(filter.shape) == 4, "Filter must be of shape (depth, height, width, channels)"
        self._filter_shape = np.array(filter.shape)
        self._convolver = Convolver(filter, dtype=dtype, norm=norm)

        shape = (int(np.prod(self._filter_shape)), int(np.prod(self._filter_shape)))
        super(RealFFTConvolve2D, self).__init__(shape=shape)

    @pycrt.enforce_precision(i="x")
    @pycu.vectorize(i="x")
    def apply(self, x: pyct.NDArray) -> pyct.NDArray:
        y = self._convolver.convolve(np.reshape(x, self._filter_shape))
        return y.ravel()

    @pycrt.enforce_precision(i="y")
    @pycu.vectorize(i="y")
    def adjoint(self, y: pyct.NDArray) -> pyct.NDArray:
        x = self._convolver.deconvolve(np.reshape(y, self._filter_shape))
        return x.ravel()


class APGD(ReconstructionAlgorithm):
    def __init__(
        self,
        psf,
        max_iter=500,
        dtype="float32",
        diff_penalty=None,
        prox_penalty=APGDPriors.NONNEG,
        acceleration=True,
        diff_lambda=0.001,
        prox_lambda=0.001,
        disp=100,
        rel_error=None,
        lipschitz_tight=True,
        lipschitz_tol=1.0,
        img_shape=None,
        **kwargs
    ):
        """
        Wrapper for `Pycsou's PGD <https://github.com/matthieumeo/pycsou/blob/a74b714192821501371c89dbd44eac15a5456a0f/src/pycsou/opt/solver/pgd.py#L17>`__
        (accelerated proximal gradient descent) applied to lensless imaging.

        Parameters
        ----------
        psf : :py:class:`~numpy.ndarray`
            Point spread function (PSF) that models forward propagation.
            Must be of shape (depth, height, width, channels) even if
            depth = 1 and channels = 1. You can use :py:func:`~lensless.io.load_psf`
            to load a PSF from a file such that it is in the correct format.
        max_iter : int, optional
            Maximal number of iterations.
        dtype : float32 or float64
            Data type to use for optimization.
        diff_penalty : None or str or `DiffFunc`
            Differentiable functional to serve as prior / regularization term.
            Default is None. See `DiffFunc <https://github.com/matthieumeo/pycsou/blob/a74b714192821501371c89dbd44eac15a5456a0f/src/pycsou/abc/operator.py#L980>`_.
        prox_penalty : None or str or `ProxFunc`
            Proximal functional to serve as prior / regularization term. Default
            is non-negative prior. See `ProxFunc <https://github.com/matthieumeo/pycsou/blob/a74b714192821501371c89dbd44eac15a5456a0f/src/pycsou/abc/operator.py#L741>`_.
        acceleration : bool, optional
            Whether to use acceleration or not. Default is True.
        diff_lambda : float
            Weight of differentiable penalty.
        prox_lambda : float
            Weight of proximal penalty.
        disp : int, optional
            Display frequency. Default is 100.
        rel_error : float, optional
            Relative error to stop optimization. Default is 1e-6.
        lipschitz_tight : bool, optional
            Whether to use tight Lipschitz constant or not. Default is True.
        lipschitz_tol : float, optional
            Tolerance to compute Lipschitz constant. Default is 1.
        img_shape : tuple, optional
            Shape of measurement (H, W, C). If None, assume shape of PSF.
        """

        assert isinstance(psf, np.ndarray), "PSF must be a numpy array"

        self._original_shape = psf.shape
        self._apgd = None

        self._stop_crit = stop.MaxIter(max_iter)
        if rel_error is not None:
            self._stop_crit = self._stop_crit | stop.RelError(eps=rel_error)
        self._disp = disp

        # Convolution (and optional downsampling) operator
        if img_shape is not None:

            meas_shape = np.array(img_shape[:2])
            rec_shape = np.array(self._original_shape[1:3])
            assert np.all(meas_shape <= rec_shape), "Image shape must be smaller than PSF shape"
            self.downsampling_factor = np.round(rec_shape / meas_shape).astype(int)

            # new PSF shape, must be integer multiple of image shape
            new_shape = tuple(np.array(meas_shape) * self.downsampling_factor) + (psf.shape[-1],)
            psf_re = resize(psf.copy(), shape=new_shape, interpolation=cv2.INTER_CUBIC)

            # combine operations
            conv = RealFFTConvolve2D(psf_re, dtype=dtype)
            ds = pycl.SubSample(
                psf_re.shape,
                slice(None),
                slice(0, -1, self.downsampling_factor[0]),
                slice(0, -1, self.downsampling_factor[1]),
                slice(None),
            )

            self._H = ds * conv

            super(APGD, self).__init__(psf_re, dtype, n_iter=max_iter, **kwargs)

        else:
            self.downsampling_factor = 1
            self._H = RealFFTConvolve2D(psf, dtype=dtype)

            super(APGD, self).__init__(psf, dtype, n_iter=max_iter, **kwargs)

        self._H.lipschitz(tol=lipschitz_tol, tight=lipschitz_tight)

        # initialize solvers which will be created when data is set
        if diff_penalty is not None:
            if diff_penalty == APGDPriors.L2:
                self._diff_penalty = diff_lambda * func.SquaredL2Norm(dim=self._H.shape[1])
            else:
                assert hasattr(diff_penalty, "jacobian")
                self._diff_penalty = diff_lambda * diff_penalty(dim=self._H.shape[1])
        else:
            self._diff_penalty = None

        if prox_penalty is not None:
            if prox_penalty == APGDPriors.L1:
                self._prox_penalty = prox_lambda * func.L1Norm(dim=self._H.shape[1])
            elif prox_penalty == APGDPriors.NONNEG:
                self._prox_penalty = prox_lambda * func.PositiveOrthant(dim=self._H.shape[1])
            else:
                try:
                    self._prox_penalty = prox_lambda * prox_penalty(dim=self._H.shape[1])
                except ValueError:
                    print("Unexpected prior.")
        else:
            self._prox_penalty = None

        self._acc = acceleration

    def set_data(self, data):
        """
        For ``APGD``, we use data to initialize problem for Pycsou.

        Parameters
        ----------
        data : :py:class:`~numpy.ndarray`
            Lensless data on which to iterate to recover an estimate of the
             scene. Should match provide PSF, i.e. shape and 2D (grayscale) or
             3D (RGB).

        """

        # super(APGD, self).set_data(
        #     np.repeat(data, self._original_shape[-4], axis=0)
        # )  # we repeat the data for each depth to match the size of the PSF

        data = np.repeat(data, self._original_shape[-4], axis=0)  # repeat for each depth
        assert isinstance(data, np.ndarray)
        assert len(data.shape) >= 3, "Data must be at least 3D: [..., width, height, channel]."

        assert np.all(
            self._psf_shape[-3:-1] == (np.array(data.shape)[-3:-1] * self.downsampling_factor)
        ), "PSF and data shape mismatch"

        if len(data.shape) == 3:
            self._data = data[None, None, ...]
        elif len(data.shape) == 4:
            self._data = data[None, ...]
        else:
            self._data = data

        """ Set up problem """
        # Cost function
        loss = (1 / 2) * func.SquaredL2Norm(dim=self._H.shape[0]).asloss(self._data.ravel())
        F = loss * self._H
        if self._diff_penalty is not None:
            F += self._diff_penalty

        self._apgd = solver.PGD(
            f=F, g=self._prox_penalty, show_progress=False, verbosity=self._disp
        )

        self._apgd.fit(
            x0=np.zeros(F.shape[1]),
            # x0=np.random.normal(size=F.shape[1]),
            stop_crit=self._stop_crit,
            track_objective=True,
            mode=pyca.solver.Mode.MANUAL,
            acceleration=self._acc,
        )

    def reset(self):
        if self._initial_est is not None:
            self._image_est = self._initial_est
        else:
            self._image_est = np.zeros(np.prod(self._psf_shape), dtype=self._dtype)

    def _update(self, iter):
        res = next(self._apgd.steps())
        self._image_est[:] = res["x"]

    def _form_image(self):
        image = self._image_est.reshape(self._psf_shape)
        image[image < 0] = 0
        if np.any(self._psf_shape != self._original_shape):
            image = resize(image, shape=self._original_shape)
        return image



"""

For Bayer data with RPI HQ sensor:
```
python scripts/measure/remote_capture.py \
rpi.username=USERNAME rpi.hostname=IP_ADDRESS
```

For Bayer data with RPI Global shutter sensor:
```
python scripts/measure/remote_capture.py -cn remote_capture_rpi_gs \
rpi.username=USERNAME rpi.hostname=IP_ADDRESS
```

For RGB data with RPI HQ RPI Global shutter sensor:
```
python scripts/measure/remote_capture.py -cn remote_capture_rpi_gs \
rpi.username=USERNAME rpi.hostname=IP_ADDRESS \
capture.bayer=False capture.down=2
```

Check out the `configs/demo.yaml` file for parameters, specifically:

- `rpi`: RPi parameters
- `capture`: parameters for taking pictures

"""

import hydra
import os
import subprocess
import cv2
from pprint import pprint
import matplotlib.pyplot as plt
from lensless.hardware.utils import check_username_hostname
from lensless.hardware.sensor import SensorOptions, sensor_dict, SensorParam
from lensless.utils.image import rgb2gray, print_image_info
from lensless.utils.plot import plot_image, pixel_histogram
from lensless.utils.io import save_image
from lensless.utils.io import load_image


@hydra.main(version_base=None, config_path="../../configs", config_name="demo")
def liveview(config):

    sensor = config.capture.sensor
    assert sensor in SensorOptions.values(), f"Sensor must be one of {SensorOptions.values()}"

    bayer = config.capture.bayer
    rgb = config.capture.rgb
    gray = config.capture.gray

    check_username_hostname(config.rpi.username, config.rpi.hostname)
    username = config.rpi.username
    hostname = config.rpi.hostname
    legacy = config.capture.legacy
    nbits_out = config.capture.nbits_out
    fn = config.capture.raw_data_fn
    gamma = config.capture.gamma
    source = config.capture.source
    plot = config.plot

    assert (
        nbits_out in sensor_dict[sensor][SensorParam.BIT_DEPTH]
    ), f"capture.nbits_out must be one of {sensor_dict[sensor][SensorParam.BIT_DEPTH]} for sensor {sensor}"
    assert (
        config.capture.nbits in sensor_dict[sensor][SensorParam.BIT_DEPTH]
    ), f"capture.nbits must be one of {sensor_dict[sensor][SensorParam.BIT_DEPTH]} for sensor {sensor}"

    if config.save:
        if config.output is not None:
            # make sure output directory exists
            os.makedirs(config.output, exist_ok=True)
            save = config.output
        else:
            save = os.getcwd()
    else:
        save = False

    # take picture
    remote_fn = "remote_capture"
    print("\nTaking picture...")
    pic_command = (
        f"{config.rpi.python} {config.capture.script} sensor={sensor} bayer={bayer} fn={remote_fn} exp={config.capture.exp} iso={config.capture.iso} "
        f"config_pause={config.capture.config_pause} sensor_mode={config.capture.sensor_mode} nbits_out={config.capture.nbits_out} "
        f"legacy={config.capture.legacy} rgb={config.capture.rgb} gray={config.capture.gray} "
    )
    if config.capture.nbits > 8:
        pic_command += " sixteen=True"
    if config.capture.down:
        pic_command += f" down={config.capture.down}"
    if config.capture.awb_gains:
        pic_command += f" awb_gains=[{config.capture.awb_gains[0]},{config.capture.awb_gains[1]}]"

    print(f"COMMAND : {pic_command}")
    ssh = subprocess.Popen(
        ["ssh", "%s@%s" % (username, hostname), pic_command],
        shell=False,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    result = ssh.stdout.readlines()
    error = ssh.stderr.readlines()

    if error != [] and legacy:  # new camera software seems to return error even if it works
        print("ERROR: %s" % error)
        return
    if result == []:
        error = ssh.stderr.readlines()
        print("ERROR: %s" % error)
        return
    else:
        result = [res.decode("UTF-8") for res in result]
        result = [res for res in result if len(res) > 3]
        result_dict = dict()
        for res in result:
            _key = res.split(":")[0].strip()
            _val = "".join(res.split(":")[1:]).strip()
            result_dict[_key] = _val
        # result_dict = dict(map(lambda s: map(str.strip, s.split(":")), result))
        print("COMMAND OUTPUT : ")
        pprint(result_dict)

    if (
        "RPi distribution" in result_dict.keys()
        and "bullseye" in result_dict["RPi distribution"]
        and not legacy
    ):
        assert not rgb or not gray, "RGB and gray not supported for RPi HQ sensor"

        if bayer:

            assert config.capture.down is None

            # copy over DNG file
            remotefile = f"~/{remote_fn}.dng"
            localfile = os.path.join(save, f"{fn}.dng")
            print(f"\nCopying over picture as {localfile}...")
            os.system('scp "%s@%s:%s" %s' % (username, hostname, remotefile, localfile))

            img = load_image(localfile, verbose=True, bayer=bayer, nbits_out=nbits_out)

            # print image properties
            print_image_info(img)

            # save as PNG
            png_out = f"{fn}.png"
            print(f"Saving RGB file as: {png_out}")
            cv2.imwrite(png_out, cv2.cvtColor(img, cv2.COLOR_RGB2BGR))

        else:

            remotefile = f"~/{remote_fn}.png"
            localfile = f"{fn}.png"
            if save:
                localfile = os.path.join(save, localfile)
            print(f"\nCopying over picture as {localfile}...")
            os.system('scp "%s@%s:%s" %s' % (username, hostname, remotefile, localfile))

            img = load_image(localfile, verbose=True)

    # legacy software running on RPi
    else:
        # copy over file
        # more pythonic? https://stackoverflow.com/questions/250283/how-to-scp-in-python
        remotefile = f"~/{remote_fn}.png"
        localfile = f"{fn}.png"
        if save:
            localfile = os.path.join(save, localfile)
        print(f"\nCopying over picture as {localfile}...")
        os.system('scp "%s@%s:%s" %s' % (username, hostname, remotefile, localfile))

        if rgb or gray:
            img = load_image(localfile, verbose=True)

        else:

            if not bayer:
                red_gain = config.camera.red_gain
                blue_gain = config.camera.blue_gain
            else:
                red_gain = None
                blue_gain = None
            # # get white balance gains
            # if red_gain is None:
            #     red_gain = float(result_dict["Red gain"])
            # if blue_gain is None:
            #     blue_gain = float(result_dict["Blue gain"])

            # # get white balance gains
            # if bayer:
            #     red_gain = 1
            #     blue_gain = 1
            # else:
            #     red_gain = float(result_dict["Red gain"])
            #     blue_gain = float(result_dict["Blue gain"])

            # load image
            print("\nLoading picture...")

            img = load_image(
                localfile,
                verbose=True,
                bayer=bayer,
                blue_gain=blue_gain,
                red_gain=red_gain,
                nbits_out=nbits_out,
            )

            # write RGB data
            if not bayer:
                cv2.imwrite(localfile, cv2.cvtColor(img, cv2.COLOR_RGB2BGR))

    # save image as viewable 8 bit
    fp = os.path.join(save, f"{fn}_8bit.png")
    save_image(img, fp, normalize=True)

    # plot RGB
    if plot:
        if not gray:
            ax = plot_image(img, gamma=gamma)
            ax.set_title("RGB")
            if save:
                plt.savefig(os.path.join(save, f"{fn}_plot.png"))

            # plot red channel
            if source == "red":
                img_1chan = img[:, :, 0]
            elif source == "green":
                img_1chan = img[:, :, 1]
            elif source == "blue":
                img_1chan = img[:, :, 2]
            else:
                img_1chan = rgb2gray(img[None, :, :, :])
            ax = plot_image(img_1chan)
            if source == "white":
                ax.set_title("Gray scale")
            else:
                ax.set_title(f"{source} channel")
            if save:
                plt.savefig(os.path.join(save, f"{fn}_1chan.png"))

            # plot histogram, useful for checking clipping
            pixel_histogram(img)
            if save:
                plt.savefig(os.path.join(save, f"{fn}_hist.png"))
            pixel_histogram(img_1chan)
            if save:
                plt.savefig(os.path.join(save, f"{fn}_1chan_hist.png"))

        else:
            ax = plot_image(img, gamma=gamma)
            if save:
                plt.savefig(os.path.join(save, f"{fn}_plot.png"))
            pixel_histogram(img)
            if save:
                plt.savefig(os.path.join(save, f"{fn}_hist.png"))

        plt.show()

    if save:
        print(f"\nSaved plots to: {save}")


if __name__ == "__main__":
    liveview()



"""
Utility file to rename files in MirFlickr25k dataset (https://press.liacs.nl/mirflickr/)
so that they match the names in the larger dataset, i.e. removing "im" from the filename.

First download MIRFLICKR-25K dataset from the above link and extract it to a directory.
```bash
wget http://press.liacs.nl/mirflickr/mirflickr25k.v3b/mirflickr25k.zip -P data/mirflickr
unzip data/mirflickr/mirflickr25k.zip -d data/mirflickr
```
Then run this script to rename the files (updating `dir_path`).
```bash
python scripts/data/rename_mirflickr25k.py
```
"""

import os
import glob
from lensless.utils.dataset import natural_sort


dir_path = "data/mirflickr/mirflickr"

# get all jpg files
files = natural_sort(glob.glob(os.path.join(dir_path, "*.jpg")))

# Rename all files in the directory
for filename in files:
    # remove "im" from the filename
    new_filename = filename.replace("im", "")

    # decrement by 1
    bn = os.path.basename(new_filename)
    file_number = int(bn.split(".")[0])
    new_bn = f"{file_number - 1}.jpg"
    new_filename = new_filename.replace(bn, new_bn)

    os.rename(filename, new_filename)

print(f"Number of files: {len(files)}")
print("Done")



"""
Capture raw Bayer data or post-processed RGB data.

```
python scripts/measure/on_device_capture.py legacy=True \
exp=0.02 bayer=True
```

With the Global Shutter sensor, legacy RPi software is not supported.
```
python scripts/measure/on_device_capture.py sensor=rpi_gs \
legacy=False exp=0.02 bayer=True
```

To capture PNG data (bayer=False) and downsample (by factor 2):
```
python scripts/measure/on_device_capture.py sensor=rpi_gs \
legacy=False exp=0.02 bayer=False down=2
```

See these code snippets for setting camera settings and post-processing
- https://github.com/scivision/pibayer/blob/1bb258c0f3f8571d6ded5491c0635686b59a5c4f/pibayer/base.py#L56
- https://picamera.readthedocs.io/en/release-1.13/recipes1.html#capturing-consistent-images
- https://www.strollswithmydog.com/open-raspberry-pi-high-quality-camera-raw

"""

import hydra
import os
import cv2
import numpy as np
from time import sleep
from PIL import Image
from lensless.hardware.utils import get_distro
from lensless.utils.image import bayer2rgb_cc, rgb2gray, resize
from lensless.hardware.constants import RPI_HQ_CAMERA_CCM_MATRIX, RPI_HQ_CAMERA_BLACK_LEVEL
from lensless.hardware.sensor import SensorOptions, sensor_dict, SensorParam
from fractions import Fraction
import time


SENSOR_MODES = [
    "off",
    "auto",
    "sunlight",
    "cloudy",
    "shade",
    "tungsten",
    "fluorescent",
    "incandescent",
    "flash",
    "horizon",
]


@hydra.main(version_base=None, config_path="../../configs", config_name="capture")
def capture(config):

    sensor = config.sensor
    assert sensor in SensorOptions.values(), f"Sensor must be one of {SensorOptions.values()}"

    bayer = config.bayer
    fn = config.fn
    exp = config.exp
    config_pause = config.config_pause
    sensor_mode = config.sensor_mode
    rgb = config.rgb
    gray = config.gray
    iso = config.iso
    sixteen = config.sixteen
    legacy = config.legacy
    down = config.down
    res = config.res
    nbits_out = config.nbits_out

    assert (
        nbits_out in sensor_dict[sensor][SensorParam.BIT_DEPTH]
    ), f"nbits_out must be one of {sensor_dict[sensor][SensorParam.BIT_DEPTH]} for sensor {sensor}"

    # https://www.raspberrypi.com/documentation/accessories/camera.html#hardware-specification
    sensor_param = sensor_dict[sensor]
    assert exp <= sensor_param[SensorParam.MAX_EXPOSURE]
    assert exp >= sensor_param[SensorParam.MIN_EXPOSURE]
    sensor_mode = int(sensor_mode)

    distro = get_distro()
    print("RPi distribution : {}".format(distro))

    if sensor == SensorOptions.RPI_GS.value:
        assert not legacy

    if "bullseye" in distro and not legacy:
        # TODO : grayscale and downsample
        assert not rgb
        assert not gray

        import subprocess

        if bayer:

            assert down is None

            # https://www.raspberrypi.com/documentation/computers/camera_software.html#raw-image-capture
            jpg_fn = fn + ".jpg"
            fn += ".dng"
            pic_command = [
                "libcamera-still",
                "-r",
                "--gain",
                f"{iso / 100}",
                "--shutter",
                f"{int(exp * 1e6)}",
                "-o",
                f"{jpg_fn}",
                # long exposure: https://www.raspberrypi.com/documentation/computers/camera_software.html#very-long-exposures
                # -- setting awbgains caused issues
                # "--awbgains 1,1",
                # "--immediate"
            ]

            cmd = subprocess.Popen(
                pic_command,
                shell=False,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            cmd.stdout.readlines()
            cmd.stderr.readlines()
            # os.remove(jpg_fn)
            os.system(f"exiftool {fn}")
            print("\nJPG saved to : {}".format(jpg_fn))
            # print("\nDNG saved to : {}".format(fn))

        else:

            from picamera2 import Picamera2, Preview

            picam2 = Picamera2()
            picam2.start_preview(Preview.NULL)

            fn += ".png"

            max_res = picam2.camera_properties["PixelArraySize"]
            if res:
                assert len(res) == 2
            else:
                res = np.array(max_res)
                if down is not None:
                    res = (np.array(res) / down).astype(int)

            res = tuple(res)
            print("Capturing at resolution: ", res)

            # capture low-dim PNG
            picam2.preview_configuration.main.size = res
            picam2.still_configuration.size = res
            picam2.still_configuration.enable_raw()
            picam2.still_configuration.raw.size = res

            # setting camera parameters
            picam2.configure(picam2.create_preview_configuration())
            new_controls = {
                "ExposureTime": int(exp * 1e6),
                "AnalogueGain": 1.0,
            }
            if config.awb_gains is not None:
                assert len(config.awb_gains) == 2
                new_controls["ColourGains"] = tuple(config.awb_gains)
            picam2.set_controls(new_controls)

            # take picture
            picam2.start("preview", show_preview=False)
            time.sleep(config.config_pause)

            picam2.switch_mode_and_capture_file("still", fn)

    # legacy camera software
    else:
        import picamerax.array

        fn += ".png"

        if bayer:

            camera = picamerax.PiCamera(framerate=1 / exp, sensor_mode=sensor_mode, resolution=res)

            # camera settings, as little processing as possible
            camera.iso = iso
            camera.shutter_speed = int(exp * 1e6)
            camera.exposure_mode = "off"
            camera.drc_strength = "off"
            camera.image_denoise = False
            camera.image_effect = "none"
            camera.still_stats = False

            sleep(config_pause)
            awb_gains = camera.awb_gains
            camera.awb_mode = "off"
            camera.awb_gains = awb_gains

            print("Resolution : {}".format(camera.resolution))
            print("Shutter speed : {}".format(camera.shutter_speed))
            print("ISO : {}".format(camera.iso))
            print("Frame rate : {}".format(camera.framerate))
            print("Sensor mode : {}".format(SENSOR_MODES[sensor_mode]))
            # keep this as it needs to be parsed from remote script!
            red_gain = float(awb_gains[0])
            blue_gain = float(awb_gains[1])
            print("Red gain : {}".format(red_gain))
            print("Blue gain : {}".format(blue_gain))

            # capture data
            stream = picamerax.array.PiBayerArray(camera)
            camera.capture(stream, "jpeg", bayer=True)

            # get bayer data
            if sixteen:
                output = np.sum(stream.array, axis=2).astype(np.uint16)
            else:
                output = (np.sum(stream.array, axis=2) >> 2).astype(np.uint8)

            # returning non-bayer data
            if rgb or gray:
                if sixteen:
                    n_bits = 12  # assuming Raspberry Pi HQ
                else:
                    n_bits = 8

                if config.awb_gains is not None:
                    red_gain = config.awb_gains[0]
                    blue_gain = config.awb_gains[1]

                output_rgb = bayer2rgb_cc(
                    output,
                    nbits=n_bits,
                    blue_gain=blue_gain,
                    red_gain=red_gain,
                    black_level=RPI_HQ_CAMERA_BLACK_LEVEL,
                    ccm=RPI_HQ_CAMERA_CCM_MATRIX,
                    nbits_out=nbits_out,
                )

                if down:
                    output_rgb = resize(
                        output_rgb[None, ...], 1 / down, interpolation=cv2.INTER_CUBIC
                    )[0]

                # need OpenCV to save 16-bit RGB image
                if gray:
                    output_gray = rgb2gray(output_rgb[None, ...])
                    output_gray = output_gray.astype(output_rgb.dtype).squeeze()
                    cv2.imwrite(fn, output_gray)
                else:
                    cv2.imwrite(fn, cv2.cvtColor(output_rgb, cv2.COLOR_RGB2BGR))
            else:
                img = Image.fromarray(output)
                img.save(fn)

        else:

            # capturing and returning non-bayer data
            from picamerax import PiCamera

            camera = PiCamera()
            if res:
                assert len(res) == 2
            else:
                res = np.array(camera.MAX_RESOLUTION)
                if down is not None:
                    res = (np.array(res) / down).astype(int)

            # -- now set up camera with desired settings
            camera = PiCamera(framerate=1 / exp, sensor_mode=sensor_mode, resolution=tuple(res))

            # Wait for the automatic gain control to settle
            time.sleep(config.config_pause)

            if config.awb_gains is not None:
                assert len(config.awb_gains) == 2
                g = (Fraction(config.awb_gains[0]), Fraction(config.awb_gains[1]))
                g = tuple(g)
                camera.awb_mode = "off"
                camera.awb_gains = g
                time.sleep(0.1)

            print("Capturing at resolution: ", res)
            print("AWB gains: ", float(camera.awb_gains[0]), float(camera.awb_gains[1]))

            try:
                camera.resolution = tuple(res)
                camera.capture(fn)
            except ValueError:
                raise ValueError(
                    "Out of resources! Use bayer for higher resolution, or increase `gpu_mem` in `/boot/config.txt`."
                )

    print("Image saved to : {}".format(fn))


if __name__ == "__main__":
    capture()



"""
To be run on the Raspberry Pi!
```
python scripts/measure/collect_dataset_on_device.py
```

Note that the script is configured for the  Raspberry Pi HQ camera

Parameters set in: configs/collect_dataset.yaml

To test on local machine, set dummy=True (which will just copy the files over).

"""
import json

import numpy as np
import hydra
from hydra.utils import to_absolute_path
import time
import os
import pathlib as plib
import shutil
import warnings
import tqdm
from picamerax import PiCamera
from fractions import Fraction
from PIL import Image
from lensless.utils.io import save_image
import re
import glob
from lensless.hardware.slm import set_programmable_mask, adafruit_sub2full

from lensless.hardware.constants import (
    RPI_HQ_CAMERA_CCM_MATRIX,
    RPI_HQ_CAMERA_BLACK_LEVEL,
)
import picamerax.array
from lensless.utils.image import bayer2rgb_cc, resize
import cv2


def convert(text):
    return int(text) if text.isdigit() else text.lower()


def alphanum_key(key):
    return [convert(c) for c in re.split("([0-9]+)", key)]


def natural_sort(arr):
    return sorted(arr, key=alphanum_key)


@hydra.main(version_base=None, config_path="../../configs/dataset", config_name="collect_dataset")
def collect_dataset(config):
    input_dir = config.input_dir
    output_dir = config.output_dir
    if output_dir is None:
        # create in same directory as input with timestamp
        output_dir = input_dir + "_measured_" + time.strftime("%Y%m%d-%H%M%S")

    MAX_TRIES = config.max_tries
    MIN_LEVEL = config.min_level
    MAX_LEVEL = config.max_level

    # if output dir exists check how many files done
    print(f"Output directory : {output_dir}")
    start_idx = config.start_idx
    if os.path.exists(output_dir):
        files = list(plib.Path(output_dir).glob(f"*.{config.output_file_ext}"))
        n_completed_files = len(files)
        print("\nNumber of completed measurements :", n_completed_files)
        output_dir = plib.Path(output_dir)
        if config.masks is not None:
            mask_dir = plib.Path(output_dir) / "masks"
    else:

        # make output directory
        output_dir = plib.Path(output_dir)
        output_dir.mkdir(exist_ok=True)

        if config.masks is not None:
            # make masks for measurements
            mask_dir = plib.Path(output_dir) / "masks"
            mask_dir.mkdir(exist_ok=True)

            np.random.seed(config.masks.seed)
            for i in range(config.masks.n):
                mask_fp = mask_dir / f"mask_{i}.npy"
                mask_vals = np.random.uniform(0, 1, config.masks.shape)
                np.save(mask_fp, mask_vals)

    recon = None
    if config.recon is not None:
        print("Initializing ADMM recon...")
        # initialize ADMM reconstruction
        from lensless import ADMM
        from lensless.utils.io import load_psf

        psf, bg = load_psf(
            fp=to_absolute_path(config.recon.psf),
            downsample=config.capture.down,  # assume full resolution PSF
            return_bg=True,
        )

        print("PSF shape: ", psf.shape)
        recon = ADMM(psf, n_iter=config.recon.n_iter)

        recon_dir = plib.Path(output_dir) / "recon"
        recon_dir.mkdir(exist_ok=True)
        print("Finished initializing ADMM recon.")

    # assert input directory exists
    assert os.path.exists(input_dir)

    # get number of files with glob
    # files = list(plib.Path(input_dir).glob(f"*.{config.input_file_ext}"))
    files = glob.glob(os.path.join(input_dir, f"*.{config.input_file_ext}"))
    files = natural_sort(files)
    files = [plib.Path(f) for f in files]
    n_files = len(files)
    print(f"\nNumber of {config.input_file_ext} files :", n_files)
    if config.n_files:
        print(f"TEST : collecting first {config.n_files} files!")
        files = files[: config.n_files]
        n_files = len(files)

    if config.runtime:
        # convert to minutes
        runtime_min = config.runtime * 60
        runtime_sec = runtime_min * 60
        if config.runtime:
            print(f"\nScript will run for (at most) {config.runtime} hour(s).")

    if config.start_delay:
        # wait for this time before starting script
        delay = config.start_delay * 60
        start_time = time.time() + delay
        start_time = time.strftime("%H:%M:%S", time.localtime(start_time))
        print(f"\nScript will start at {start_time}")
        time.sleep(delay)

    print("\nStarting measurement!\n")
    start_time = time.time()

    if not config.dummy:

        res = config.capture.res
        down = config.capture.down

        # set up camera for consistent photos
        # https://picamera.readthedocs.io/en/release-1.13/recipes1.html#capturing-consistent-images
        # https://picamerax.readthedocs.io/en/latest/fov.html?highlight=camera%20resolution#sensor-modes
        # -- just get max resolution of camera
        camera = PiCamera(framerate=30)
        if res:
            assert len(res) == 2
        else:
            res = np.array(camera.MAX_RESOLUTION)
            if down is not None:
                res = (np.array(res) / down).astype(int)
        camera.close()

        # -- now set up camera with desired settings
        max_increase = (
            config.capture.fact_increase * config.max_tries
            if config.max_tries > 0
            else 1
        )
        max_exposure = min(20, config.capture.exposure * max_increase)
        if config.capture.framerate is None:
            framerate = 1 / max_exposure
            warnings.warn(
                f"Framerate is not given. Setting it to 1 / max_exposure = {framerate}"
            )
        elif config.capture.framerate > 1 / max_exposure:
            warnings.warn(
                f"Framerate should be less or equal 1 / max_exposure = {1 / max_exposure}. Resetting framerate"
            )
            framerate = 1 / max_exposure
        else:
            framerate = config.capture.framerate

        camera = PiCamera(sensor_mode=0, resolution=tuple(res), framerate=framerate)

        # Set ISO to the desired value
        camera.resolution = tuple(res)
        camera.iso = config.capture.iso
        # Wait for the automatic gain control to settle
        time.sleep(config.capture.config_pause)
        # Now fix the values

        if config.capture.exposure:
            # in microseconds
            init_shutter_speed = int(config.capture.exposure * 1e6)
        else:
            init_shutter_speed = camera.exposure_speed
        camera.exposure_mode = "off"
        camera.shutter_speed = init_shutter_speed

        # AWB
        if config.capture.awb_gains:
            assert len(config.capture.awb_gains) == 2
            g = (Fraction(config.capture.awb_gains[0]), Fraction(config.capture.awb_gains[1]))
            g = tuple(g)
        else:
            g = camera.awb_gains

        camera.awb_mode = "off"
        camera.awb_gains = g

        # for parameters to settle
        time.sleep(config.capture.config_pause)

        print("Capturing at resolution: ", res)
        print("AWB gains", float(camera.awb_gains[0]), float(camera.awb_gains[1]))

    init_brightness = config.display.brightness

    # loop over files with tqdm
    exposure_vals = []
    brightness_vals = []
    n_tries_vals = []

    bg_name = None
    current_bg = {}
    shutter_speed = init_shutter_speed
    for i, _file in enumerate(tqdm.tqdm(files[start_idx:])):

        # save file in output directory as PNG
        output_fp = output_dir / _file.name
        output_fp = output_fp.with_suffix(f".{config.output_file_ext}")

        # if not done, perform measurement
        if not os.path.isfile(output_fp):
            if config.dummy:
                shutil.copyfile(_file, output_fp)
                time.sleep(1)
            else:

                # display img
                display_img(_file, config, init_brightness)
                # capture img
                output, _, _, camera = capture_screen(
                    MAX_LEVEL,
                    MAX_TRIES,
                    MIN_LEVEL,
                    _file,
                    brightness_vals,
                    camera,
                    config,
                    down,
                    exposure_vals,
                    g,
                    i,
                    init_brightness,
                    shutter_speed,
                    None,
                    n_tries_vals,
                    output_fp,
                    start_idx,
                )

                if config.capture.measure_bg:
                    # name of background for current image
                    bg_name = plib.Path(config.capture.bg_fp + str(i)).with_suffix(
                        f".{config.output_file_ext}"
                    )
                    bg = output_dir / bg_name

                    # append current file to bg list
                    if str(bg_name) not in current_bg:
                        current_bg[str(bg_name)] = str(_file.name)
                    else:
                        current_bg[str(bg_name)].append(str(_file.name))
                    # capture background periodically
                    if i % config.capture.measure_bg == 0 or (i == n_files - 1):
                        bg_name = str(bg_name)
                        # push the last bg-capture pairs
                        if current_bg:
                            with open(output_dir / "bg_mappings.json", "a") as outfile:
                                json.dump(current_bg, outfile, indent=4)
                        current_bg = {}
                        # current_bg[bg_name] = None

                        # display bg
                        display_img(None, config, brightness=init_brightness)
                        # capture bg
                        output, shutter_speed, init_brightness, camera = capture_screen(
                            MAX_LEVEL,
                            0,
                            MIN_LEVEL,
                            plib.Path(config.capture.bg_fp + str(i)).with_suffix(
                                f".{config.output_file_ext}"
                            ),
                            brightness_vals,
                            camera,
                            config,
                            down,
                            exposure_vals,
                            g,
                            i,
                            init_brightness,
                            shutter_speed,
                            None,
                            n_tries_vals,
                            bg,
                            start_idx,
                        )

        if recon is not None:
            # normalize and remove background
            output = output.astype(np.float32)
            output /= output.max()
            output -= bg  # TODO implement fancy bg subtraction
            output = np.clip(output, a_min=0, a_max=output.max())

            # set data
            output = output[np.newaxis, :, :, :]
            recon.set_data(output)

            # reconstruct and save
            res = recon.apply()
            recon_fp = recon_dir / output_fp.name
            save_image(res, recon_fp)

        # check if runtime is exceeded
        if config.runtime:
            proc_time = time.time() - start_time
            if proc_time > runtime_sec:
                print(f"-- measured {i + 1} / {n_files} files")
                break

    proc_time = time.time() - start_time
    print(f"\nFinished, {proc_time / 60.:.3f} minutes.")

    # print brightness and exposure range and average
    print(f"brightness range: {np.min(brightness_vals)} - {np.max(brightness_vals)}")
    print(f"exposure range: {np.min(exposure_vals)} - {np.max(exposure_vals)}")
    print(f"n_tries range: {np.min(n_tries_vals)} - {np.max(n_tries_vals)}")
    print(f"brightness average: {np.mean(brightness_vals)}")
    print(f"exposure average: {np.mean(exposure_vals)}")
    print(f"n_tries average: {np.mean(n_tries_vals)}")


def capture_screen(
    MAX_LEVEL,
    MAX_TRIES,
    MIN_LEVEL,
    _file,
    brightness_vals,
    camera,
    config,
    down,
    exposure_vals,
    g,
    i,
    init_brightness,
    init_shutter_speed,
    mask_dir,
    n_tries_vals,
    output_fp,
    start_idx,
):
    if not config.capture.skip:

        # -- set mask pattern
        if config.masks is not None:
            mask_idx = (i + start_idx) % config.masks.n
            mask_fp = mask_dir / f"mask_{mask_idx}.npy"
            print("using mask: ", mask_fp)
            mask_vals = np.load(mask_fp)
            full_pattern = adafruit_sub2full(
                mask_vals,
                center=config.masks.center,
            )
            set_programmable_mask(full_pattern, device=config.masks.device)

        # -- take picture
        max_pixel_val = 0
        fact_increase = config.capture.fact_increase
        fact_decrease = config.capture.fact_decrease
        n_tries = 0

        if MAX_TRIES > 0:
            # shutter speed is constant for MAX_TRIES == 0
            camera.shutter_speed = int(init_shutter_speed)
            time.sleep(config.capture.config_pause)
        current_shutter_speed = camera.shutter_speed

        current_screen_brightness = init_brightness
        print(f"current shutter speed: {current_shutter_speed}")
        print(f"current screen brightness: {current_screen_brightness}")

        while max_pixel_val < MIN_LEVEL or max_pixel_val > MAX_LEVEL:

            # get bayer data
            stream = picamerax.array.PiBayerArray(camera)
            camera.capture(stream, "jpeg", bayer=True)
            output_bayer = np.sum(stream.array, axis=2).astype(np.uint16)

            # convert to RGB
            output = bayer2rgb_cc(
                output_bayer,
                down=down,
                nbits=12,
                blue_gain=float(g[1]),
                red_gain=float(g[0]),
                black_level=RPI_HQ_CAMERA_BLACK_LEVEL,
                ccm=RPI_HQ_CAMERA_CCM_MATRIX,
                nbits_out=8,
            )

            # if down:
            #     output = resize(
            #         output[None, ...], factor=1 / down, interpolation=cv2.INTER_CUBIC
            #     )[0]

            # save image
            save_image(output, output_fp, normalize=False)

            # print range
            print(f"{output_fp}, range: {output.min()} - {output.max()}")

            n_tries += 1
            if n_tries > MAX_TRIES:
                if MAX_TRIES != 0:
                    print("Max number of tries reached!")
                break

            max_pixel_val = output.max()

            if max_pixel_val < MIN_LEVEL:

                # increase exposure
                current_shutter_speed = int(current_shutter_speed * fact_increase)
                camera.shutter_speed = current_shutter_speed
                time.sleep(config.capture.config_pause)

                print(
                    f"increasing shutter speed to [desired] {current_shutter_speed} [actual] {camera.shutter_speed}"
                )

            elif max_pixel_val > MAX_LEVEL:
                if current_shutter_speed > 13098:  # TODO: minimum for RPi HQ
                    # decrease exposure
                    current_shutter_speed = int(current_shutter_speed / fact_decrease)
                    camera.shutter_speed = current_shutter_speed
                    time.sleep(config.capture.config_pause)
                    print(
                        f"decreasing shutter speed to [desired] {current_shutter_speed} [actual] {camera.shutter_speed}"
                    )

                else:

                    # decrease screen brightness
                    current_screen_brightness = current_screen_brightness - 10
                    display_img(_file, config, current_screen_brightness)

        exposure_vals.append(current_shutter_speed / 1e6)
        brightness_vals.append(current_screen_brightness)
        n_tries_vals.append(n_tries)
    return output, current_shutter_speed, current_screen_brightness, camera


def display_img(_file, config, brightness):
    if _file is None:
        point_source = np.zeros(tuple(config.display.screen_res) + (3,))
        fp = "tmp_display.png"
        im = Image.fromarray(point_source.astype("uint8"), "RGB")
        im.save(fp)
        _file = fp

    # -- show on display
    screen_res = np.array(config.display.screen_res)
    hshift = config.display.hshift
    vshift = config.display.vshift
    pad = config.display.pad

    display_image_path = config.display.output_fp
    rot90 = config.display.rot90
    display_command = f"python scripts/measure/prep_display_image.py --fp {_file} --output_path {display_image_path} --screen_res {screen_res[0]} {screen_res[1]} --hshift {hshift} --vshift {vshift} --pad {pad} --brightness {brightness} --rot90 {rot90}"
    if config.display.landscape:
        display_command += " --landscape"
    if config.display.image_res is not None:
        display_command += (
            f" --image_res {config.display.image_res[0]} {config.display.image_res[1]}"
        )
    # print(display_command)
    os.system(display_command)
    time.sleep(config.display.delay)


if __name__ == "__main__":
    collect_dataset()



"""
Push dataset measured with LenslessPiCam to HuggingFace.

```bash
# install
pip install datasets
pip install huggingface_hub

# make a write token on HuggingFace

# run
python scripts/data/upload_dataset_huggingface.py \
hf_token=... \
```
"""

import hydra
from hydra.utils import to_absolute_path
import time
import os
import glob
import numpy as np
from datasets import Dataset, DatasetDict, Image
from huggingface_hub import upload_file
from lensless.utils.dataset import natural_sort
from tqdm import tqdm
from lensless.utils.io import save_image
import cv2
from joblib import Parallel, delayed


@hydra.main(
    version_base=None, config_path="../../configs/dataset", config_name="upload_dataset_huggingface"
)
def upload_dataset(config):

    start_time = time.time()

    # parameters
    repo_id = config.repo_id
    hf_token = config.hf_token
    n_files = config.n_files
    multimask = config.multimask
    n_jobs = config.n_jobs
    assert hf_token is not None, "Please provide a HuggingFace token."
    assert repo_id is not None, "Please provide a HuggingFace repo_id."
    assert config.lensless.dir is not None, "Please provide a lensless directory."
    assert config.lensed.dir is not None, "Please provide a lensed directory."
    assert (
        config.lensless.ext is not None
    ), "Please provide a lensless file extension, e.g. .png, .jpg, .tiff"
    assert (
        config.lensed.ext is not None
    ), "Please provide a lensed file extension, e.g. .png, .jpg, .tiff"

    config.lensless.dir = to_absolute_path(config.lensless.dir)
    config.lensed.dir = to_absolute_path(config.lensed.dir)

    # get masks
    files_masks = []
    n_masks = 0
    if multimask:
        files_masks = glob.glob(os.path.join(config.lensless.dir, "masks", "*.npy"))
        files_masks = natural_sort(files_masks)
        n_masks = len(files_masks)

    # get lensless files
    files_lensless = glob.glob(os.path.join(config.lensless.dir, "*" + config.lensless.ext))
    files_lensless = natural_sort(files_lensless)
    print(f"Number of lensless files: {len(files_lensless)}")
    if n_files is not None:
        print(f"Only keeping {n_files} files...")
        files_lensless = files_lensless[:n_files]

    # get lensed files
    files_lensed = glob.glob(os.path.join(config.lensed.dir, "*" + config.lensed.ext))
    print(f"Number of lensed files: {len(files_lensed)}")

    # only keep if in both
    bn_lensless = [os.path.basename(f).split(".")[0] for f in files_lensless]
    bn_lensed = [os.path.basename(f).split(".")[0] for f in files_lensed]
    common_files = list(set(bn_lensless).intersection(bn_lensed))
    common_files = natural_sort(common_files)
    print(f"Number of common files: {len(common_files)}")

    # get file paths
    lensless_files = [
        os.path.join(config.lensless.dir, f + config.lensless.ext) for f in common_files
    ]
    lensed_files = [os.path.join(config.lensed.dir, f + config.lensed.ext) for f in common_files]
    background_files = []
    if config.lensless.ambient:
        # check that corresponding ambient files exist
        for f in common_files:
            ambient_bn = "black_background" + os.path.basename(f) + config.lensless.ext
            ambient_f = os.path.join(config.lensless.dir, ambient_bn)
            assert os.path.exists(ambient_f), f"File {ambient_f} does not exist."
            background_files.append(ambient_f)

    if config.lensless.downsample is not None:

        if config.lensless.ambient:
            raise NotImplementedError("Downsampling not implemented for ambient files.")

        tmp_dir = config.lensless.dir + "_tmp"
        os.makedirs(tmp_dir, exist_ok=True)

        def downsample(f, output_dir):
            img = cv2.imread(f, cv2.IMREAD_UNCHANGED)
            img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
            img = cv2.resize(
                img,
                (0, 0),
                fx=1 / config.lensless.downsample,
                fy=1 / config.lensless.downsample,
                interpolation=cv2.INTER_LINEAR,
            )
            new_fp = os.path.join(output_dir, os.path.basename(f))
            new_fp = new_fp.split(".")[0] + config.lensless.ext
            save_image(img, new_fp, normalize=False)

        Parallel(n_jobs=n_jobs)(delayed(downsample)(f, tmp_dir) for f in tqdm(lensless_files))
        lensless_files = glob.glob(os.path.join(tmp_dir, f"*{config.lensless.ext[1:]}"))

    # convert to normalized 8 bit
    if config.lensless.eight_norm:

        if config.lensless.ambient:
            raise NotImplementedError("Normalized 8-bit not implemented for ambient files.")

        tmp_dir = config.lensless.dir + "_tmp"
        os.makedirs(tmp_dir, exist_ok=True)

        # -- parallelize with joblib
        def save_8bit(f, output_dir, normalize=True):
            img = cv2.imread(f, cv2.IMREAD_UNCHANGED)
            img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
            new_fp = os.path.join(output_dir, os.path.basename(f))
            new_fp = new_fp.split(".")[0] + config.lensless.ext
            save_image(img, new_fp, normalize=normalize)

        Parallel(n_jobs=n_jobs)(delayed(save_8bit)(f, tmp_dir) for f in tqdm(lensless_files))
        lensless_files = glob.glob(os.path.join(tmp_dir, f"*{config.lensless.ext[1:]}"))

    # check for attribute
    df_attr = None
    if "celeba_attr" in config.lensed.keys():
        # load attribute txt file with pandas
        import pandas as pd

        fp = config.lensed.celeba_attr
        df = pd.read_csv(fp, sep=r"\s+", header=1, index_col=0)
        df_attr = df[: len(common_files)]
        # convert -1 to 0
        df_attr = df_attr.replace(-1, 0)
        # convert to boolean
        df_attr = df_attr.astype(bool)
        # to dict
        df_attr = df_attr.to_dict(orient="list")
    if multimask:
        # add label according to mask
        mask_labels = [np.arange(n_masks)]
        mask_labels = np.tile(mask_labels, (len(common_files), 1))
        mask_labels = mask_labels.flatten()
        mask_labels = mask_labels[: len(common_files)]
        if df_attr is not None:
            df_attr["mask_label"] = mask_labels
        else:
            df_attr = {"mask_label": mask_labels}

    # step 1: create Dataset objects
    def create_dataset(lensless_files, lensed_files, df_attr=None, ambient_files=None):
        dataset_dict = {
            "lensless": lensless_files,
            "lensed": lensed_files,
        }
        if df_attr is not None:
            # combine dictionaries
            dataset_dict = {**dataset_dict, **df_attr}
        if ambient_files is not None:
            dataset_dict["ambient"] = ambient_files
        dataset = Dataset.from_dict(dataset_dict)
        dataset = dataset.cast_column("lensless", Image())
        dataset = dataset.cast_column("lensed", Image())
        if ambient_files is not None:
            dataset = dataset.cast_column("ambient", Image())
        return dataset

    # train-test split
    test_size = config.test_size
    if multimask:
        n_mask_test = int(test_size * n_masks)
        # get indices from df_attr
        test_indices = np.where(df_attr["mask_label"] < n_mask_test)[0]
        train_indices = np.where(df_attr["mask_label"] >= n_mask_test)[0]
        # split dict into train-test
        test_dataset = create_dataset(
            [lensless_files[i] for i in test_indices],
            [lensed_files[i] for i in test_indices],
            {k: [v[i] for i in test_indices] for k, v in df_attr.items()},
            ambient_files=[background_files[i] for i in test_indices]
            if config.lensless.ambient
            else None,
        )
        train_dataset = create_dataset(
            [lensless_files[i] for i in train_indices],
            [lensed_files[i] for i in train_indices],
            {k: [v[i] for i in train_indices] for k, v in df_attr.items()},
            ambient_files=[background_files[i] for i in train_indices]
            if config.lensless.ambient
            else None,
        )
    elif isinstance(config.split, int):
        n_test_split = int(test_size * config.split)

        # get all indices
        n_splits = np.ceil(len(lensless_files) / config.split).astype(int)
        test_idx = np.array([])
        for i in range(n_splits):
            test_idx = np.append(test_idx, np.arange(n_test_split) + i * config.split)
        test_idx = test_idx.astype(int)

        # get train indices
        train_idx = np.setdiff1d(np.arange(len(lensless_files)), test_idx)
        train_idx = train_idx.astype(int)

        # split dict into train-test
        test_dataset = create_dataset(
            [lensless_files[i] for i in test_idx],
            [lensed_files[i] for i in test_idx],
            ambient_files=[background_files[i] for i in test_idx]
            if config.lensless.ambient
            else None,
        )
        train_dataset = create_dataset(
            [lensless_files[i] for i in train_idx],
            [lensed_files[i] for i in train_idx],
            ambient_files=[background_files[i] for i in train_idx]
            if config.lensless.ambient
            else None,
        )

    else:
        n_test = int(test_size * len(common_files))
        if df_attr is not None:
            # split dict into train-test
            df_attr_test = {k: v[:n_test] for k, v in df_attr.items()}
            df_attr_train = {k: v[n_test:] for k, v in df_attr.items()}
        else:
            df_attr_test = None
            df_attr_train = None
        test_dataset = create_dataset(
            lensless_files[:n_test],
            lensed_files[:n_test],
            df_attr_test,
            ambient_files=background_files[:n_test] if config.lensless.ambient else None,
        )
        train_dataset = create_dataset(
            lensless_files[n_test:],
            lensed_files[n_test:],
            df_attr_train,
            ambient_files=background_files[n_test:] if config.lensless.ambient else None,
        )
    print(f"Train size: {len(train_dataset)}")
    print(f"Test size: {len(test_dataset)}")

    # step 2: create DatasetDict
    dataset_dict = DatasetDict(
        {
            "train": train_dataset,
            "test": test_dataset,
        }
    )

    # step 3: push to hub
    if config.files is not None:
        for f in config.files:
            fp = to_absolute_path(config.files[f])
            ext = os.path.splitext(fp)[1]
            remote_fn = f"{f}{ext}"
            upload_file(
                path_or_fileobj=fp,
                path_in_repo=remote_fn,
                repo_id=repo_id,
                repo_type="dataset",
                token=hf_token,
            )

            # viewable version of file if it is an image
            if ext in [".png", ".jpg", ".jpeg", ".tiff"]:
                img = cv2.imread(fp, cv2.IMREAD_UNCHANGED)
                local_fp = f"{f}_viewable8bit.png"
                remote_fn = f"{f}_viewable8bit.png"
                save_image(img, local_fp, normalize=True)
                upload_file(
                    path_or_fileobj=local_fp,
                    path_in_repo=remote_fn,
                    repo_id=repo_id,
                    repo_type="dataset",
                    token=hf_token,
                )

    dataset_dict.push_to_hub(repo_id, token=hf_token)

    upload_file(
        path_or_fileobj=lensless_files[0],
        # path_in_repo=f"lensless_example{config.lensless.ext}" if not config.lensless.eight_norm else f"lensless_example.png",
        path_in_repo=f"lensless_example{config.lensless.ext}",
        repo_id=repo_id,
        repo_type="dataset",
        token=hf_token,
    )
    upload_file(
        path_or_fileobj=lensed_files[0],
        path_in_repo=f"lensed_example{config.lensed.ext}",
        repo_id=repo_id,
        repo_type="dataset",
        token=hf_token,
    )
    if config.lensless.ambient:
        upload_file(
            path_or_fileobj=background_files[0],
            path_in_repo=f"ambient_example{config.lensless.ext}",
            repo_id=repo_id,
            repo_type="dataset",
            token=hf_token,
        )

    for _mask_file in files_masks:
        upload_file(
            path_or_fileobj=_mask_file,
            path_in_repo=os.path.join("masks", os.path.basename(_mask_file)),
            repo_id=repo_id,
            repo_type="dataset",
            token=hf_token,
        )

    # total time in minutes
    print(f"Total time: {(time.time() - start_time) / 60} minutes")

    # delete PNG files
    if config.lensless.eight_norm or config.lensless.downsample:
        os.system(f"rm -rf {tmp_dir}")


if __name__ == "__main__":
    upload_dataset()



# #############################################################################
# admm.py
# =================
# Authors :
# Eric BEZZAM [ebezzam@gmail.com]
# Julien SAHLI [julien.sahli@epfl.ch]
# #############################################################################


import numpy as np
from lensless.recon.recon import ReconstructionAlgorithm
from scipy import fft
from lensless.utils.io import load_data
import time

try:
    import torch

    torch_available = True
except ImportError:
    torch_available = False


class ADMM(ReconstructionAlgorithm):
    """
    Object for applying ADMM (Alternating Direction Method of Multipliers) with
    a non-negativity constraint and a total variation (TV) prior.

    Paper about ADMM: https://web.stanford.edu/~boyd/papers/pdf/admm_distr_stats.pdf

    Slides about ADMM: https://web.stanford.edu/class/ee364b/lectures/admm_slides.pdf

    """

    def __init__(
        self,
        psf,
        dtype=None,
        mu1=1e-6,
        mu2=1e-5,
        mu3=4e-5,
        tau=0.0001,
        psi=None,
        psi_adj=None,
        psi_gram=None,
        pad=False,
        norm="backward",
        # PnP
        denoiser=None,
        **kwargs,
    ):
        """

        Parameters
        ----------
        psf : :py:class:`~numpy.ndarray` or :py:class:`~torch.Tensor`
            Point spread function (PSF) that models forward propagation.
            Must be of shape (depth, height, width, channels) even if
            depth = 1 and channels = 1. You can use :py:func:`~lensless.io.load_psf`
            to load a PSF from a file such that it is in the correct format.
        dtype : float32 or float64
            Data type to use for optimization. Default is float32.
        mu1 : float
            Step size for updating primal/dual variables.
        mu2 : float
            Step size for updating primal/dual variables.
        mu3 : float
            Step size for updating primal/dual variables.
        tau : float
            Weight for L1 norm of `psi` applied to the image estimate.
        psi : :py:class:`function`, optional
            Operator to map image to a space that the image is assumed to be
            sparse in (hence L1 norm). Default is to use total variation (TV)
            operator.
        psi_adj : :py:class:`function`
            Adjoint of `psi`.
        psi_gram : :py:class:`function`
            Function to compute gram of `psi`.
        pad : bool
            Whether to pad the image with zeros before applying the PSF. Default
            is False, as optimized data is already padded.
        norm : str
            Normalization to use for the convolution. Options are "forward",
            "backward", and "ortho". Default is "backward".
        """
        self._mu1 = mu1
        self._mu2 = mu2
        self._mu3 = mu3
        self._tau = tau

        # 3D ADMM is not supported yet
        assert len(psf.shape) == 4, "PSF must be 4D: (depth, height, width, channels)."
        if psf.shape[0] > 1:
            raise NotImplementedError(
                "3D ADMM is not supported yet, use gradient descent or APGD instead."
            )

        # call reset() to initialize matrices
        self._proj = self._Psi
        super(ADMM, self).__init__(
            psf, dtype, pad=pad, norm=norm, denoiser=denoiser, reset=False, **kwargs
        )

        # set prior
        if psi is None:
            # use already defined Psi and PsiT
            self._PsiTPsi = finite_diff_gram(self._padded_shape, self._dtype, self.is_torch)
        else:
            assert psi_adj is not None
            assert psi_gram is not None
            assert callable(psi)
            assert callable(psi_adj)
            assert callable(psi_gram)
            # overwrite already defined Psi and PsiT
            self._Psi = psi
            self._PsiT = psi_adj
            self._PsiTPsi = psi_gram(self._padded_shape)

            # - need to reset with new projector
            self._proj = self._Psi

        # precompute_R_divmat (self._H computed by constructor with reset())
        if self.is_torch:
            self._PsiTPsi = self._PsiTPsi.to(self._psf.device)

        # check denoiser for PnP
        if self._denoiser is not None:
            self._denoiser_use_dual = denoiser["use_dual"]

            # - need to reset with new projector
            self._proj = self._denoiser
            # identify function
            self._PsiT = lambda x: x

        self.reset()

    def _Psi(self, x):
        """
        Operator to map image to space that the image is assumed to be sparse
        in.
        """
        return finite_diff(x)

    def _PsiT(self, U):
        """
        Adjoint of `_Psi`.
        """
        return finite_diff_adj(U)

    def reset(self):
        if self.is_torch:
            # TODO initialize without padding
            # initialize image estimate as [Batch, Depth, Height, Width, Channels]
            if self._initial_est is not None:
                self._image_est = self._initial_est
            else:
                self._image_est = torch.zeros([1] + self._padded_shape, dtype=self._dtype).to(
                    self._psf.device
                )

            # self._image_est = torch.zeros_like(self._psf)
            self._X = torch.zeros_like(self._image_est)
            # self._U = torch.zeros_like(self._Psi(self._image_est))
            if self._denoiser is not None:
                # PnP
                self._U = torch.zeros_like(
                    self._denoiser(self._image_est, self._denoiser_noise_level)
                )
            else:
                self._U = torch.zeros_like(self._proj(self._image_est))
            self._W = torch.zeros_like(self._X)
            if self._image_est.max():
                # if non-zero
                # self._forward_out = self._forward()
                self._forward_out = self._convolver.convolve(self._image_est)
                self._Psi_out = self._Psi(self._image_est)
            else:
                self._forward_out = torch.zeros_like(self._X)
                self._Psi_out = torch.zeros_like(self._U)

            self._xi = torch.zeros_like(self._image_est)
            self._eta = torch.zeros_like(self._U)
            self._rho = torch.zeros_like(self._X)

            # precompute _R_divmat
            self._R_divmat = 1.0 / (
                self._mu1 * (torch.abs(self._convolver._Hadj * self._convolver._H))
                + self._mu2 * torch.abs(self._PsiTPsi)
                + self._mu3
            ).type(self._complex_dtype)

            # precompute_X_divmat
            self._X_divmat = 1.0 / (self._convolver._pad(torch.ones_like(self._psf)) + self._mu1)
            # self._X_divmat = 1.0 / (torch.ones_like(self._psf) + self._mu1)

        else:
            if self._initial_est is not None:
                self._image_est = self._initial_est
            else:
                self._image_est = np.zeros([1] + self._padded_shape, dtype=self._dtype)

            # self._U = np.zeros(np.r_[self._padded_shape, [2]], dtype=self._dtype)
            self._X = np.zeros_like(self._image_est)
            # self._U = np.zeros_like(self._Psi(self._image_est))
            self._U = np.zeros_like(self._proj(self._image_est))
            self._W = np.zeros_like(self._X)
            if self._image_est.max():
                # if non-zero
                # self._forward_out = self._forward()
                self._forward_out = self._convolver.convolve(self._image_est)
                self._Psi_out = self._Psi(self._image_est)
            else:
                self._forward_out = np.zeros_like(self._X)
                self._Psi_out = np.zeros_like(self._U)

            self._xi = np.zeros_like(self._image_est)
            self._eta = np.zeros_like(self._U)
            self._rho = np.zeros_like(self._X)

            # precompute R_divmat
            self._R_divmat = 1.0 / (
                self._mu1 * (np.abs(self._convolver._Hadj * self._convolver._H))
                + self._mu2 * np.abs(self._PsiTPsi)
                + self._mu3
            ).astype(self._complex_dtype)

            # precompute_X_divmat
            self._X_divmat = 1.0 / (
                self._convolver._pad(np.ones(self._psf_shape, dtype=self._dtype)) + self._mu1
            )

    def _U_update(self):
        """Total variation update."""
        # to avoid computing sparse operator twice
        if self._denoiser is not None:
            # PnP
            if self._denoiser_use_dual:
                self._U = self._denoiser(
                    self._U + self._eta / self._mu2,
                    self._denoiser_noise_level,
                )
            else:
                self._U = self._denoiser(self._image_est, self._denoiser_noise_level)
        else:
            self._U = soft_thresh(
                self._Psi_out + self._eta / self._mu2, thresh=self._tau / self._mu2
            )

    def _X_update(self):
        # to avoid computing forward model twice
        # self._X = self._X_divmat * (self._xi + self._mu1 * self._forward_out + self._data)
        self._X = self._X_divmat * (
            self._xi + self._mu1 * self._forward_out + self._convolver._pad(self._data)
        )

    def _W_update(self):
        """Non-negativity update"""
        if self.is_torch:
            self._W = torch.maximum(
                self._rho / self._mu3 + self._image_est, torch.zeros_like(self._image_est)
            )
        else:
            self._W = np.maximum(self._rho / self._mu3 + self._image_est, 0)

    def _image_update(self):
        if self._denoiser is not None:
            # PnP
            rk = (
                (self._mu3 * self._W - self._rho)
                # + self._mu2 * self._U
                + self._mu2 * self._U - self._eta
                if self._denoiser_use_dual
                else self._mu2 * self._U
                + self._convolver.deconvolve(self._mu1 * self._X - self._xi)
            )
        else:
            rk = (
                (self._mu3 * self._W - self._rho)
                + self._PsiT(self._mu2 * self._U - self._eta)
                + self._convolver.deconvolve(self._mu1 * self._X - self._xi)
            )

        # rk = self._convolver._pad(rk)

        if self.is_torch:
            freq_space_result = self._R_divmat * torch.fft.rfft2(rk, dim=(-3, -2))
            self._image_est = torch.fft.irfft2(
                freq_space_result, dim=(-3, -2), s=self._convolver._padded_shape[-3:-1]
            )
        else:
            freq_space_result = self._R_divmat * fft.rfft2(rk, axes=(-3, -2))
            self._image_est = fft.irfft2(
                freq_space_result, axes=(-3, -2), s=self._convolver._padded_shape[-3:-1]
            )

        # self._image_est = self._convolver._crop(res)

    def _xi_update(self):
        # to avoid computing forward model twice
        self._xi += self._mu1 * (self._forward_out - self._X)

    def _eta_update(self):
        # to avoid finite difference operataion again?
        if self._denoiser is not None:
            # PnP
            self._eta += self._mu2 * (self._image_est - self._U)
        else:
            self._eta += self._mu2 * (self._Psi_out - self._U)

    def _rho_update(self):
        self._rho += self._mu3 * (self._image_est - self._W)

    def _update(self, iter):
        self._U_update()
        self._X_update()
        self._W_update()
        self._image_update()

        # update forward and sparse operators
        self._forward_out = self._convolver.convolve(self._image_est)
        if self._denoiser is None:
            self._Psi_out = self._Psi(self._image_est)

        self._xi_update()
        if self._denoiser is None:
            self._eta_update()
        elif self._denoiser_use_dual:
            self._eta_update()
        self._rho_update()

    def _form_image(self):
        image = self._convolver._crop(self._image_est)

        # # TODO without cropping
        # image = self._image_est

        image[image < 0] = 0
        return image


def soft_thresh(x, thresh):
    if torch_available and isinstance(x, torch.Tensor):
        return torch.sign(x) * torch.max(torch.abs(x) - thresh, torch.zeros_like(x))
    else:
        # numpy automatically applies functions to each element of the array
        return np.sign(x) * np.maximum(0, np.abs(x) - thresh)


def finite_diff(x):
    """Gradient of image estimate, approximated by finite difference. Space where image is assumed sparse."""
    if torch_available and isinstance(x, torch.Tensor):
        return torch.stack(
            (torch.roll(x, 1, dims=-3) - x, torch.roll(x, 1, dims=-2) - x), dim=len(x.shape)
        )
    else:
        return np.stack(
            (np.roll(x, 1, axis=-3) - x, np.roll(x, 1, axis=-2) - x),
            axis=len(x.shape),
        )


def finite_diff_adj(x):
    """Adjoint of finite difference operator."""
    if torch_available and isinstance(x, torch.Tensor):
        diff1 = torch.roll(x[..., 0], -1, dims=-3) - x[..., 0]
        diff2 = torch.roll(x[..., 1], -1, dims=-2) - x[..., 1]
    else:
        diff1 = np.roll(x[..., 0], -1, axis=-3) - x[..., 0]
        diff2 = np.roll(x[..., 1], -1, axis=-2) - x[..., 1]
    return diff1 + diff2


def finite_diff_gram(shape, dtype=None, is_torch=False):
    """Gram matrix of finite difference operator."""
    if is_torch:
        if dtype is None:
            dtype = torch.float32
        gram = torch.zeros(shape, dtype=dtype)

    else:
        if dtype is None:
            dtype = np.float32
        gram = np.zeros(shape, dtype=dtype)

    if shape[0] == 1:
        gram[0, 0, 0] = 4
        gram[0, 0, 1] = gram[0, 0, -1] = gram[0, 1, 0] = gram[0, -1, 0] = -1
    else:
        gram[0, 0, 0] = 6
        gram[0, 0, 1] = gram[0, 0, -1] = gram[0, 1, 0] = gram[0, -1, 0] = gram[1, 0, 0] = gram[
            -1, 0, 0
        ] = -1

    if is_torch:
        return torch.fft.rfft2(gram, dim=(-3, -2))
    else:
        return fft.rfft2(gram, axes=(-3, -2))


def apply_admm(psf_fp, data_fp, n_iter, verbose=False, **kwargs):

    # load data
    psf, data = load_data(psf_fp=psf_fp, data_fp=data_fp, plot=False, **kwargs)

    # create reconstruction object
    recon = ADMM(psf, n_iter=n_iter)

    # set data
    recon.set_data(data)

    # perform reconstruction
    start_time = time.time()
    res = recon.apply(plot=False)
    proc_time = time.time() - start_time

    if verbose:
        print(f"Reconstruction time : {proc_time} s")
        print(f"Reconstruction shape: {res.shape}")
    return res



"""

Simulate a mask, simulate a measurement with it, and reconstruct the image.

Procedure is as follows:

1) Simulate the mask.
2) Simulate a measurement with the mask and specified physical parameters.
3) Reconstruct the image from the measurement.

Example usage:

Simulate FlatCam with separable simulation and Tikhonov reconstuction (https://arxiv.org/abs/1509.00116, Eq 7):
```
# MLS mask
python scripts/sim/mask_single_file.py mask.type=MLS simulation.flatcam=True recon.algo=tikhonov

# MURA mask
python scripts/sim/mask_single_file.py mask.type=MURA mask.n_bits=99 simulation.flatcam=True recon.algo=tikhonov
```

Using Torch
```
python scripts/sim/mask_single_file.py mask.type=MLS simulation.flatcam=True recon.algo=tikhonov use_torch=True
```

Simulate FlatCam with PSF simulation and Tikhonov reconstuction:
```
python scripts/sim/mask_single_file.py mask.type=MLS simulation.flatcam=False recon.algo=tikhonov
```

Simulate FlatCam with PSF simulation and ADMM reconstruction. Doesn't work well.
```
python scripts/sim/mask_single_file.py mask.type=MLS simulation.flatcam=False recon.algo=admm
```

Simulate Fresnel Zone Aperture camera with PSF simulation and ADMM reconstuction (https://www.nature.com/articles/s41377-020-0289-9):
Doesn't work well, maybe need to remove DC offset which hurts reconstructions?
```
python scripts/sim/mask_single_file.py mask.type=FZA recon.algo=admm recon.admm.n_iter=18
```

Simulate PhaseContour camera with PSF simulation and ADMM reconstuction (https://ieeexplore.ieee.org/document/9076617):
```
python scripts/sim/mask_single_file.py mask.type=PhaseContour recon.algo=admm recon.admm.n_iter=10
```

"""

import hydra
import warnings
from hydra.utils import to_absolute_path
from lensless.utils.io import load_image, save_image
from lensless.utils.image import rgb2gray, rgb2bayer, bayer2rgb
import numpy as np
import matplotlib.pyplot as plt
from lensless import ADMM
from lensless.utils.plot import plot_image
from lensless.eval.metric import mse, psnr, ssim, lpips
from waveprop.simulation import FarFieldSimulator
import os
from lensless.hardware.mask import CodedAperture, PhaseContour, FresnelZoneAperture
from lensless.recon.tikhonov import CodedApertureReconstruction
import torch


@hydra.main(version_base=None, config_path="../../configs/simulate", config_name="mask_sim_single")
def simulate(config):

    fp = to_absolute_path(config.files.original)
    assert os.path.exists(fp), f"File {fp} does not exist."

    # simulation parameters
    object_height = config.simulation.object_height
    scene2mask = config.simulation.scene2mask
    mask2sensor = config.simulation.mask2sensor
    sensor = config.simulation.sensor
    snr_db = config.simulation.snr_db
    downsample = config.simulation.downsample
    max_val = config.simulation.max_val

    image_format = config.simulation.image_format.lower()
    if image_format not in ["grayscale", "rgb"]:
        bayer = True
    else:
        bayer = False

    # 1) simulate mask
    mask_type = config.mask.type
    if mask_type.upper() in ["MURA", "MLS"]:
        mask = CodedAperture.from_sensor(
            sensor_name=sensor,
            downsample=downsample,
            method=mask_type,
            distance_sensor=mask2sensor,
            **config.mask,
        )
    elif mask_type.upper() == "FZA":
        mask = FresnelZoneAperture.from_sensor(
            sensor_name=sensor,
            downsample=downsample,
            distance_sensor=mask2sensor,
            **config.mask,
        )
    elif mask_type.lower() == "PhaseContour".lower():
        mask = PhaseContour.from_sensor(
            sensor_name=sensor,
            downsample=downsample,
            distance_sensor=mask2sensor,
            n_iter=config.mask.phase_mask_iter,
            **config.mask,
        )

    # 2) simulate measurement
    image = load_image(fp, verbose=True) / 255
    if config.use_torch:
        image = image.transpose(2, 0, 1)
        image = torch.from_numpy(image).float()

    flatcam_sim = config.simulation.flatcam
    if flatcam_sim and mask_type.upper() not in ["MURA", "MLS"]:
        warnings.warn(
            "Flatcam simulation only supported for MURA and MLS masks. Using far field simulation with PSF."
        )
        flatcam_sim = False

    # use far field simulator to get correct object plane sizing
    psf = mask.psf
    if config.use_torch:
        psf = psf.transpose(2, 0, 1)
        psf = torch.from_numpy(psf).float()

    simulator = FarFieldSimulator(
        psf=psf,
        object_height=object_height,
        scene2mask=scene2mask,
        mask2sensor=mask2sensor,
        sensor=sensor,
        snr_db=snr_db,
        max_val=max_val,
        is_torch=config.use_torch,
    )
    image_plane, object_plane = simulator.propagate(image, return_object_plane=True)

    # channels as last dimension
    if config.use_torch:
        image_plane = image_plane.permute(1, 2, 0)
        object_plane = object_plane.permute(1, 2, 0)
        image = image.permute(1, 2, 0)

    if image_format == "grayscale":
        image_plane = rgb2gray(image_plane)
        object_plane = rgb2gray(object_plane)
    elif "bayer" in image_format:
        image_plane = rgb2bayer(image_plane, pattern=image_format[-4:])
        object_plane = rgb2bayer(object_plane, pattern=image_format[-4:])
    else:
        # make sure image is RGB
        assert image_plane.shape[-1] == 3, "Image plane must be RGB"
        assert object_plane.shape[-1] == 3, "Object plane must be RGB"

    if flatcam_sim:
        image_plane = mask.simulate(object_plane, snr_db=snr_db)

    # 3) reconstruct image
    save = config["save"]
    if save:
        save = os.getcwd()

    if config.recon.algo.lower() == "tikhonov":
        recon = CodedApertureReconstruction(
            mask, object_plane.shape, lmbd=config.recon.tikhonov.reg
        )
        recovered = recon.apply(image_plane)

    elif config.recon.algo.lower() == "admm":

        if bayer:
            raise ValueError("ADMM reconstruction not supported for Bayer images.")

        # prep PSF
        if image_format == "grayscale":
            psf = rgb2gray(mask.psf)
        else:
            psf = mask.psf
        psf = psf[np.newaxis, :, :, :] / np.linalg.norm(mask.psf.ravel())

        # prep recon
        recon = ADMM(psf, **config.recon.admm)

        # add depth channel
        recon.set_data(image_plane[None, :, :, :])
        res = recon.apply(
            n_iter=config.recon.admm.n_iter, disp_iter=config.recon.admm.disp_iter, save=save
        )[0]

        # remove depth channel
        recovered = res[0]
    else:
        raise ValueError(f"Reconstruction algorithm {config.recon.algo} not recognized.")

    # back to numpy for evaluation and plotting
    if config.use_torch:
        recovered = recovered.numpy()
        object_plane = object_plane.numpy()
        image_plane = image_plane.numpy()

    # 4) evaluate
    if image_format == "grayscale":
        object_plane = object_plane[:, :, 0]
        recovered = recovered[:, :, 0]

    print("\nEvaluation:")
    print("MSE", mse(object_plane, recovered))
    print("PSNR", psnr(object_plane, recovered))
    if image_format == "grayscale":
        print("SSIM", ssim(object_plane, recovered, channel_axis=None))
    else:
        print("SSIM", ssim(object_plane, recovered))
    if image_format == "rgb":
        print("LPIPS", lpips(object_plane, recovered))

    # -- plot
    if bayer:
        print("Converting to RGB for plotting and saving...")
        image_plane = bayer2rgb(image_plane, pattern=image_format[-4:])
        object_plane = bayer2rgb(object_plane, pattern=image_format[-4:])
        recovered = bayer2rgb(recovered, pattern=image_format[-4:])

    _, ax = plt.subplots(ncols=5, nrows=1, figsize=(24, 5))
    plot_image(object_plane, ax=ax[0])
    ax[0].set_title("Object plane")
    if np.iscomplexobj(mask.mask):
        # plot phase
        plot_image(np.angle(mask.mask), ax=ax[1])
        ax[1].set_title("Phase mask")
    else:
        plot_image(mask.mask, ax=ax[1])
        ax[1].set_title("Amplitude mask")
    plot_image(mask.psf, ax=ax[2], gamma=2.2)
    ax[2].set_title("PSF")
    plot_image(image_plane, ax=ax[3])
    ax[3].set_title("Raw data")
    plot_image(recovered, ax=ax[4])
    ax[4].set_title("Reconstruction")

    for a in ax:
        a.set_axis_off()

    plt.tight_layout()
    plt.savefig("result.png")

    if config.save:
        save_image(recovered, "reconstruction.png")

    plt.show()


if __name__ == "__main__":
    simulate()
